from ....posix.libgen import Basename, Dirname
from ..systemv import MIPSSysVModel


class MIPSSysVBasename(Basename, MIPSSysVModel):
    pass


class MIPSSysVDirname(Dirname, MIPSSysVModel):
    pass


__all__ = [
    "MIPSSysVBasename",
    "MIPSSysVDirname",
]



import typing

from .... import emulators
from ....platforms import Architecture
from ..c99.signal import Signal
from ..cstd import ArgumentType, CStdModel

# NOTE: signal() and raise() are defined in C99

# NOTE: pid_t is defined as int in all supported ABIs.
# NOTE: pthread_t is defined as unsigned long in all supported ABIs.

# NOTE: glibc and POSIX actively disagree about sigmask_t
# POSIX headers define sigmask_t as a 128-byte type.
# glibc only uses the first eight bytes;
# it doesn't touch the trailing 120 bytes.


class BsdSignal(Signal):
    # NOTE: This is signal() with slightly different semantics.
    #
    # The C99 specification made some rather unilateral decisions
    # regarding what should happen when a signal handler fires.
    #
    # The BSD community disagreed with the System V community,
    # and made their own unilateral decisions
    # which are enshrined in the API for this function.
    #
    # The System V community fired back with the non-standard
    # sysv_signal() function.
    #
    # POSIX fixed this by introducing sigaction(),
    # which lets you control all of these decisions for yourself.
    #
    # This is why there are warnings everywhere saying
    # "don't use signal() except with pre-defined handlers".
    name = "bsd_signal"

    # typedef void (*sighandler_t)(int);
    # sighandler_t bsd_signal(int sig, sighandler_t func);


class Kill(CStdModel):
    name = "kill"

    # int kill(pid_t pid, int sig);
    argument_types = [ArgumentType.INT, ArgumentType.INT]
    return_type = ArgumentType.INT

    # Most of our emulators can't model signals.
    # I'm not sure it's possible to model raising one.
    imprecise = True

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        # Not really modeled; just return success
        self.set_return_value(emulator, 0)


class Killpg(CStdModel):
    name = "killpg"

    # int killpg(int pgid, int sig);
    argument_types = [ArgumentType.INT, ArgumentType.INT]
    return_type = ArgumentType.INT

    # Most of our emulators can't model signals.
    # I'm not sure it's possible to model raising one.
    imprecise = True

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        # Not really modeled; just return success
        self.set_return_value(emulator, 0)


class PthreadKill(CStdModel):
    name = "pthread_kill"

    # int pthread_kill(pthread_t thread, int sig);
    argument_types = [ArgumentType.ULONG, ArgumentType.INT]
    return_type = ArgumentType.INT

    # Most of our emulators can't model signals.
    # I'm not sure it's possible to model raising one.
    # We currently have no plans to model multithreading.
    imprecise = True

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        # Not really modeled; just return success
        self.set_return_value(emulator, 0)


class PthreadSigmask(CStdModel):
    name = "pthread_sigmask"

    # int pthread_sigmask(int how, sigset_t *set, sigset_t *oldset);
    argument_types = [ArgumentType.INT, ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    # We currently have no plans to model multithreading.
    # This model assumes a single thread,
    # since I have no concept of a current TID.
    imprecise = True

    # NOTE: This assumes a null initial signal mask.

    # TODO: This won't interoperate with sigprocmask()
    # You're not supposed to interoperate the two,
    # so I'm okay with this for now.

    def __init__(self, address: int):
        super().__init__(address)
        self.sigmask = b"\0" * 128

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        how = self.get_arg1(emulator)
        newset = self.get_arg2(emulator)
        oldset = self.get_arg3(emulator)

        assert isinstance(how, int)
        assert isinstance(newset, int)
        assert isinstance(oldset, int)

        # sigset_t is a 1024-bit bit vector.
        # It's split up differently in different ABIs, but that's what it amounts to.

        # TODO: This syscall actually handles out-of-bounds addresses
        # Do we want to model this, or leave the error to let harness authors know?

        if oldset != 0:
            # If oldset is set, save the sigmask to it.
            emulator.write_memory(oldset, self.sigmask)

        if newset != 0:
            # Read the new vector out of memory.
            # If this thing has alignment issues, I will make ugly noises.
            newvec = emulator.read_memory(newset, 128)

            # OF COURSE MIPS IS DIFFERENT!
            is_mips = (
                1
                if self.platform.architecture
                in (Architecture.MIPS32, Architecture.MIPS64)
                else 0
            )

            if how == 0 + is_mips:
                # Case: how == SIG_BLOCK: Set sigmask to newvec | sigmask
                self.sigmask = bytes(
                    map(lambda x: self.sigmask[x] | newvec[x], range(0, 128))
                )
            elif how == 1 + is_mips:
                # Case: how == SIG_UNBLOCK: Set sigmask to ~newvec | sigmask

                self.sigmask = bytes(
                    map(lambda x: self.sigmask[x] & ~newvec[x], range(0, 128))
                )
            elif how == 2 + is_mips:
                # Case: how == SIG_SETMASK: Set sigmask to newvec
                self.sigmask = newvec
            else:
                # Case: default: Return failure
                print(f"Unknown how {how}")
                self.set_return_value(emulator, -1)
                return

        self.set_return_value(emulator, 0)


class Sigaction(CStdModel):
    name = "sigaction"

    # typedef void (*sighandler_t)(int);
    # int sigaction(int sig, const struct sigaction *act, const struct sigaction *oldact);
    argument_types = [ArgumentType.INT, ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    # FIXME: This doesn't interoperate with signal()
    # Programs generally shouldn't use signal(), but they can use it to clear a handler.

    # Most of our emulators can't model signals
    imprecise = True

    def __init__(self, address: int):
        super().__init__(address)
        # Track any registered handlers.
        # Sometimes programs care about this.
        self.handlers: typing.Dict[int, int] = dict()
        self.flags: typing.Dict[int, int] = dict()
        self.masks: typing.Dict[int, bytes] = dict()

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        sig = self.get_arg1(emulator)
        act = self.get_arg2(emulator)
        oldact = self.get_arg3(emulator)

        assert isinstance(sig, int)
        assert isinstance(act, int)
        assert isinstance(oldact, int)

        # Figure out field alignments in the sigaction struct.
        # Of course MIPS has to be different.
        if self.platform.architecture == Architecture.MIPS32:
            handler_off = 0x4
            mask_off = 0x8
            flags_off = 0x0
        elif self.platform.architecture == Architecture.MIPS64:
            handler_off = 0x8
            mask_off = 0x10
            flags_off = 0x0
        elif ArgumentType.LONG in self._four_byte_types:
            handler_off = 0x0
            mask_off = 0x4
            flags_off = 0x84
        else:
            handler_off = 0x0
            mask_off = 0x8
            flags_off = 0x88

        # Get the old handler details out of this model
        oldhandler = self.handlers.get(sig, 0)
        oldmask = self.masks.get(sig, b"\0" * 128)
        oldflags = self.flags.get(sig, 0)

        if act != 0:
            # If act is non-null, set the handler to its current contents
            newhandler = self.read_integer(
                act + handler_off, ArgumentType.POINTER, emulator
            )
            newmask = emulator.read_memory(act + mask_off, 128)
            newflags = self.read_integer(act + flags_off, ArgumentType.UINT, emulator)

            self.handlers[sig] = newhandler
            self.masks[sig] = newmask
            self.flags[sig] = newflags

        if oldact != 0:
            # If oldact is non-null, populate it with the old handler info
            self.write_integer(
                oldact + handler_off, oldhandler, ArgumentType.POINTER, emulator
            )
            emulator.write_memory(oldact + mask_off, oldmask)
            self.write_integer(
                oldact + flags_off, oldflags, ArgumentType.UINT, emulator
            )

        self.set_return_value(emulator, 0)


class Sigaddset(CStdModel):
    name = "sigaddset"

    # int sigaddset(sigset_t *set, int sig);
    argument_types = [ArgumentType.POINTER, ArgumentType.INT]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        ptr = self.get_arg1(emulator)
        idx = self.get_arg2(emulator)

        assert isinstance(ptr, int)
        assert isinstance(idx, int)

        # NOTE: This uses glibc's model of how this works.
        # This only supports 64 signals for most platforms,
        # 128 for MIPS.
        max_idx = 64
        if self.platform.architecture in (Architecture.MIPS32, Architecture.MIPS64):
            max_idx = 128

        # Signal numbers are one-indexed
        idx -= 1

        if idx < 0 or idx >= max_idx:
            # Bad signal number
            self.set_return_value(emulator, -1)
            return

        if ArgumentType.LONG in self._four_byte_types:
            bitoff = idx % 32
            wordoff = idx // 32
            wordsize = 4
        else:
            bitoff = idx % 64
            wordoff = idx // 64
            wordsize = 8

        # Read the specific word storing our bit.
        # The array is LSW-first, but the words are in native byte order.
        # Technically they're longs, but it's easier to handle them as ulongs.
        word = self.read_integer(
            ptr + (wordoff * wordsize), ArgumentType.ULONG, emulator
        )
        word |= 1 << bitoff
        self.write_integer(
            ptr + (wordoff * wordsize), word, ArgumentType.ULONG, emulator
        )

        self.set_return_value(emulator, 0)


class Sigaltstack(CStdModel):
    name = "sigaltstack"

    # int sigaltstack(const stack_t *stack, stack_t *oldstack);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    # All none of the modeled signals will use the alternate stack.
    imprecise = True

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        # Not really modeled; just return success
        self.set_return_value(emulator, 0)


class Sigdelset(CStdModel):
    name = "sigdelset"

    # int sigdelset(sigset_t *set, int sig);
    argument_types = [ArgumentType.POINTER, ArgumentType.INT]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        ptr = self.get_arg1(emulator)
        idx = self.get_arg2(emulator)

        assert isinstance(ptr, int)
        assert isinstance(idx, int)

        # NOTE: This uses glibc's model of how this works.
        # This only supports 64 signals for most platforms,
        # 128 for MIPS.
        max_idx = 64
        if self.platform.architecture in (Architecture.MIPS32, Architecture.MIPS64):
            max_idx = 128

        # Signal numbers are one-indexed
        idx -= 1

        if idx < 0 or idx >= max_idx:
            # Bad signal number
            self.set_return_value(emulator, -1)
            return

        if ArgumentType.LONG in self._four_byte_types:
            bitoff = idx % 32
            wordoff = idx // 32
            wordsize = 4
        else:
            bitoff = idx % 64
            wordoff = idx // 64
            wordsize = 8

        # Read the specific word storing our bit.
        # The array is LSW-first, but the words are in native byte order.
        # Technically they're longs, but it's easier to handle them as ulongs.
        word = self.read_integer(
            ptr + (wordoff * wordsize), ArgumentType.ULONG, emulator
        )
        word &= ~(1 << bitoff)
        self.write_integer(
            ptr + (wordoff * wordsize), word, ArgumentType.ULONG, emulator
        )

        self.set_return_value(emulator, 0)


class Sigemptyset(CStdModel):
    name = "sigemptyset"

    # int sigemptyset(sigset_t *set);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        ptr = self.get_arg1(emulator)

        assert isinstance(ptr, int)

        # NOTE: This follows the GNU standard.
        # GNU only uses the first few bytes of a sigset_t.
        emulator.write_memory(ptr, b"\0" * 8)

        # Of course MIPS is different
        if self.platform.architecture in (Architecture.MIPS32, Architecture.MIPS64):
            emulator.write_memory(ptr + 8, b"\0" * 8)

        self.set_return_value(emulator, 0)


class Sigfillset(CStdModel):
    name = "sigfillset"

    # int sigfillset(sigset_t *set);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        ptr = self.get_arg1(emulator)

        assert isinstance(ptr, int)

        # NOTE: This follows the GNU standard.
        # GNU only uses the first few bytes of a sigset_t.
        # It also excludes two real-time signals.

        if ArgumentType.LONG in self._four_byte_types:
            # I'd love to use ulonglong, but the word ordering isn't endian-dependent.
            self.write_integer(ptr, 0x7FFFFFFF, ArgumentType.ULONG, emulator)
            self.write_integer(ptr + 4, 0xFFFFFFFE, ArgumentType.ULONG, emulator)
        else:
            self.write_integer(ptr, 0xFFFFFFFE7FFFFFFF, ArgumentType.ULONG, emulator)

        # Of course MIPS is different.
        if self.platform.architecture in (Architecture.MIPS32, Architecture.MIPS64):
            # All bits are set, so it doesn't matter how I write this.
            self.write_integer(
                ptr + 8, 0xFFFFFFFFFFFFFFFF, ArgumentType.ULONGLONG, emulator
            )

        self.set_return_value(emulator, 0)


class Sighold(CStdModel):
    name = "sighold"

    # int sighold(int sig);
    argument_types = [ArgumentType.INT]
    return_type = ArgumentType.INT

    # This is part of a historical System V signal handler.
    # Unlike sigaction and sigprocmask,
    # it's not easy to localize across functions.
    imprecise = True

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        # Not really modeled; just return success
        self.set_return_value(emulator, 0)


class Sigignore(CStdModel):
    name = "sigignore"

    # int sigignore(int sig);
    argument_types = [ArgumentType.INT]
    return_type = ArgumentType.INT

    # This is part of a historical System V signal handler.
    # Unlike sigaction and sigprocmask,
    # it's not easy to localize across functions.
    imprecise = True

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        # Not really modeled; just return success
        self.set_return_value(emulator, 0)


class Siginterrupt(CStdModel):
    name = "siginterrupt"

    # int siginterrupt(int sig, int flag);
    argument_types = [ArgumentType.INT, ArgumentType.INT]
    return_type = ArgumentType.INT

    # This is part of a historical System V signal handler.
    # Unlike sigaction and sigprocmask,
    # it's not easy to localize across functions.
    imprecise = True

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        # Not really modeled; just return success
        self.set_return_value(emulator, 0)


class Sigismember(CStdModel):
    name = "sigismember"

    # int sigismember(sigset_t *set, int sig);
    argument_types = [ArgumentType.POINTER, ArgumentType.INT]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)


class Sigpause(CStdModel):
    name = "sigpause"

    # int sigpause(int arg);
    argument_types = [ArgumentType.POINTER, ArgumentType.INT]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        # This one's fun.  If it never gets signaled, it never returns.
        raise NotImplementedError(f"No idea how to correctly model {self.name}")


class Sigpending(Sigemptyset):
    name = "sigpending"

    # int sigpending(sigset_t *set);

    # We don't support signals.
    # If you need pending signals, this will be a problem
    imprecise = True

    # Since we don't support signals, no signals will ever be pending.
    # This is semantically equivalent to sigemptyset()


class Sigprocmask(PthreadSigmask):
    name = "sigprocmask"

    # int sigprocmask(int, const sigset_t *set, sigset_t *oldset);

    # Without threads, this is identical to PthreadSigmask


class Sigqueue(CStdModel):
    name = "sigqueue"

    # int sigqueue(pid_t pid, int sig, const union sigval);
    argument_types = [ArgumentType.INT, ArgumentType.INT, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    # This doesn't actually send a signal
    imprecise = True

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        # Not really modeled; just return success
        self.set_return_value(emulator, 0)


class Sigrelse(CStdModel):
    name = "sigrelse"

    # int sigrelse(int sig);
    argument_types = [ArgumentType.INT]
    return_type = ArgumentType.INT

    # This is part of a historical System V signal handler.
    # Unlike sigaction and sigprocmask,
    # it's not easy to localize across functions.
    imprecise = True

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        # Not really modeled; just return success
        self.set_return_value(emulator, 0)


class Sigset(CStdModel):
    name = "sigset"

    # typedef void (*sighandler_t)(int);
    # sighandler_t sigset(int, sighandler_t handler);
    argument_types = [ArgumentType.INT, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    # This is part of a historical System V signal handler.
    # Unlike sigaction and sigprocmask,
    # it's not easy to localize across functions.
    imprecise = True

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        # Not really modeled; just return success
        self.set_return_value(emulator, 0)


class Sigsuspend(CStdModel):
    name = "sigsuspend"

    # int sigsuspend(sigset_t *set);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        # This one's fun.  If it never gets signaled, it never returns.
        raise NotImplementedError(f"No idea how to correctly model {self.name}")


class Sigtimedwait(CStdModel):
    name = "sigtimedwait"

    # int sigtimedwait(const sigset_t *set, siginfo_t *info, const struct timespec *timespec);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    # This model assumes it always times out,
    # even if no timeout was specified
    imprecise = True

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        # Not really modeled; just return.
        # TODO: This API relies on errno
        self.set_return_value(emulator, -1)


class Sigwait(CStdModel):
    name = "sigwait"

    # int sigwait(const sigset_t *set, int *sig);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        # This one's fun.  If it never gets signaled, it never returns.
        raise NotImplementedError(f"No idea how to correctly model {self.name}")


class Sigwaitinfo(CStdModel):
    name = "sigwaitinfo"

    # int sigwaitinfo(const sigset_t *set, siginfo_t *sig);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        # This one's fun.  If it never gets signaled, it never returns.
        raise NotImplementedError(f"No idea how to correctly model {self.name}")


__all__ = [
    "BsdSignal",
    "Kill",
    "Killpg",
    "PthreadKill",
    "PthreadSigmask",
    "Sigaction",
    "Sigaddset",
    "Sigaltstack",
    "Sigdelset",
    "Sigemptyset",
    "Sigfillset",
    "Sighold",
    "Sigignore",
    "Siginterrupt",
    "Sigismember",
    "Sigpause",
    "Sigpending",
    "Sigprocmask",
    "Sigqueue",
    "Sigrelse",
    "Sigset",
    "Sigsuspend",
    "Sigtimedwait",
    "Sigwait",
    "Sigwaitinfo",
]



import locale

from .... import emulators, exceptions
from ..cstd import ArgumentType, CStdModel
from ..errno import ErrnoResolver
from .utils import (
    MAX_STRLEN,
    _emu_memcmp,
    _emu_memcpy,
    _emu_strlen,
    _emu_strncat,
    _emu_strncmp,
    _emu_strncpy,
    _emu_strnlen,
)


class Memcpy(CStdModel):
    name = "memcpy"

    # void *memcpy(void *restrict dst, const void *restrict src, size_t n);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER, ArgumentType.SIZE_T]
    return_type = ArgumentType.POINTER

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        dst = self.get_arg1(emulator)
        src = self.get_arg2(emulator)
        n = self.get_arg3(emulator)

        assert isinstance(dst, int)
        assert isinstance(src, int)
        assert isinstance(n, int)

        # FIXME: Does not actually mimic memcpy
        # Will not clobber overlapping buffers
        _emu_memcpy(emulator, dst, src, n)
        self.set_return_value(emulator, dst)


class Memmove(CStdModel):
    name = "memmove"

    # void *memmove(void *restrict dst, const void *restrict src, size_t n);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER, ArgumentType.SIZE_T]
    return_type = ArgumentType.POINTER

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        dst = self.get_arg1(emulator)
        src = self.get_arg2(emulator)
        n = self.get_arg3(emulator)

        assert isinstance(dst, int)
        assert isinstance(src, int)
        assert isinstance(n, int)

        _emu_memcpy(emulator, dst, src, n)
        self.set_return_value(emulator, dst)


class Strcpy(CStdModel):
    name = "strcpy"

    # char *strcpy(char *dst, const char *src);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.POINTER

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        dst = self.get_arg1(emulator)
        src = self.get_arg2(emulator)

        assert isinstance(dst, int)
        assert isinstance(src, int)

        n = _emu_strlen(emulator, src) + 1

        _emu_memcpy(emulator, dst, src, n)
        self.set_return_value(emulator, dst)


class Strncpy(CStdModel):
    name = "strncpy"

    # char *strcpy(char *dst, const char *src, size_t n);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER, ArgumentType.SIZE_T]
    return_type = ArgumentType.POINTER

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        dst = self.get_arg1(emulator)
        src = self.get_arg2(emulator)
        n = self.get_arg3(emulator)

        assert isinstance(dst, int)
        assert isinstance(src, int)
        assert isinstance(n, int)

        _emu_strncpy(emulator, dst, src, n)
        self.set_return_value(emulator, dst)


class Strcat(CStdModel):
    name = "strcat"

    # char *strcat(char *restrict s1, const char *restrict s2);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.POINTER

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        dst = self.get_arg1(emulator)
        src = self.get_arg2(emulator)

        assert isinstance(dst, int)
        assert isinstance(src, int)

        _emu_strncat(emulator, dst, src, MAX_STRLEN)
        self.set_return_value(emulator, dst)


class Strncat(CStdModel):
    name = "strncat"

    # char *strncat(char *restrict s1, const char *restrict s2, size_t n);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER, ArgumentType.SIZE_T]
    return_type = ArgumentType.POINTER

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        dst = self.get_arg1(emulator)
        src = self.get_arg2(emulator)
        n = self.get_arg3(emulator)

        assert isinstance(dst, int)
        assert isinstance(src, int)
        assert isinstance(n, int)

        _emu_strncat(emulator, dst, src, n)
        self.set_return_value(emulator, dst)


class Memcmp(CStdModel):
    name = "memcmp"

    # int memcmp(const void *ptr1, const void *ptr2, size_t n);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER, ArgumentType.SIZE_T]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        ptr1 = self.get_arg1(emulator)
        ptr2 = self.get_arg2(emulator)
        n = self.get_arg3(emulator)

        assert isinstance(ptr1, int)
        assert isinstance(ptr2, int)
        assert isinstance(n, int)

        res = _emu_memcmp(emulator, ptr1, ptr2, n)
        self.set_return_value(emulator, res)


class Strncmp(CStdModel):
    name = "strncmp"

    # int strncmp(const void *ptr1, const void *ptr2, size_t n);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER, ArgumentType.SIZE_T]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        # int strncmp(const char *ptr1, const char *ptr2, size_t n);
        ptr1 = self.get_arg1(emulator)
        ptr2 = self.get_arg2(emulator)
        n = self.get_arg3(emulator)

        assert isinstance(ptr1, int)
        assert isinstance(ptr2, int)
        assert isinstance(n, int)

        res = _emu_strncmp(emulator, ptr1, ptr2, n)
        self.set_return_value(emulator, res)


class Strcmp(CStdModel):
    name = "strcmp"

    # int strcmp(const void *ptr1, const void *ptr2);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        # int strcmp(const char *ptr1, const char *ptr2);
        ptr1 = self.get_arg1(emulator)
        ptr2 = self.get_arg2(emulator)

        assert isinstance(ptr1, int)
        assert isinstance(ptr2, int)

        res = _emu_strncmp(emulator, ptr1, ptr2, MAX_STRLEN)
        self.set_return_value(emulator, res)


class Strcoll(CStdModel):
    name = "strcoll"

    # int strcoll(const void *ptr1, const void *ptr2);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.INT

    # Won't respond to dynamic changes to the locale.
    imprecise = True

    def __init__(self, address: int):
        super().__init__(address)
        # NOTE: This requries extra configuration; set `locale` to the preferred locale.
        # TODO: Think of a way to support dynamically-changing locales.
        self.locale = ""

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        ptr1 = self.get_arg1(emulator)
        ptr2 = self.get_arg2(emulator)

        assert isinstance(ptr1, int)
        assert isinstance(ptr2, int)

        # TODO: This might be wrong if CTYPE is different
        len1 = _emu_strlen(emulator, ptr1)
        len2 = _emu_strlen(emulator, ptr2)

        bytes1 = emulator.read_memory(ptr1, len1)
        bytes2 = emulator.read_memory(ptr2, len2)

        try:
            # Risky.  Inside this section, the locale will be different.
            old_locale = locale.getlocale(category=locale.LC_COLLATE)
            locale.setlocale(locale.LC_COLLATE, self.locale)
            locale.setlocale(locale.LC_CTYPE, self.locale)

            encoding = locale.getpreferredencoding()

            str1 = bytes1.decode(encoding)
            str2 = bytes2.decode(encoding)

            res = locale.strcoll(str1, str2)

            locale.setlocale(locale.LC_COLLATE, old_locale)
            locale.setlocale(locale.LC_CTYPE, old_locale)
        except Exception as e:
            locale.setlocale(locale.LC_COLLATE, old_locale)
            locale.setlocale(locale.LC_CTYPE, old_locale)
            raise e

        self.set_return_value(emulator, res)


class Strxfrm(CStdModel):
    name = "strxfrm"

    # size_t strxfrm(char *dst, const char *src, size_t n);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER, ArgumentType.SIZE_T]
    return_type = ArgumentType.SIZE_T

    # Won't respond to dynamic changes to the locale.
    imprecise = True

    def __init__(self, address: int):
        super().__init__(address)
        # NOTE: This requries extra configuration; set `locale` to the preferred locale.
        # TODO: Think of a way to support dynamically-changing locales.
        self.locale = ""

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        dst = self.get_arg1(emulator)
        src = self.get_arg2(emulator)
        n = self.get_arg3(emulator)

        assert isinstance(dst, int)
        assert isinstance(src, int)
        assert isinstance(n, int)

        # TODO: This might be wrong if CTYPE is different
        if n == 0:
            n = _emu_strlen(emulator, src)
        else:
            n = _emu_strnlen(emulator, src, n)

        self.set_return_value(emulator, n)

        if dst == 0:
            return

        bytes1 = emulator.read_memory(src, n)

        try:
            # Risky.  Inside this section, the locale will be different.
            old_locale = locale.getlocale(category=locale.LC_COLLATE)
            locale.setlocale(locale.LC_COLLATE, self.locale)
            locale.setlocale(locale.LC_CTYPE, self.locale)

            encoding = locale.getpreferredencoding()

            str1 = bytes1.decode(encoding)

            str2 = locale.strxfrm(str1)

            bytes2 = str2.encode(encoding)

            locale.setlocale(locale.LC_COLLATE, old_locale)
            locale.setlocale(locale.LC_CTYPE, old_locale)
        except Exception as e:
            locale.setlocale(locale.LC_COLLATE, old_locale)
            locale.setlocale(locale.LC_CTYPE, old_locale)
            raise e

        emulator.write_memory(dst, bytes2)


class Memchr(CStdModel):
    name = "memchr"

    # const void *memchr(const void *ptr, int value, size_t n);
    argument_types = [ArgumentType.POINTER, ArgumentType.INT, ArgumentType.SIZE_T]
    return_type = ArgumentType.SIZE_T

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        # const void *memchr(const void *ptr, int value, size_t n);
        ptr = self.get_arg1(emulator)
        val = self.get_arg2(emulator)
        n = self.get_arg3(emulator)

        assert isinstance(ptr, int)
        assert isinstance(val, int)
        assert isinstance(n, int)

        data = emulator.read_memory(ptr, n)
        for i in range(0, n):
            if data[i] == val:
                self.set_return_value(emulator, ptr + i)
                return

        self.set_return_value(emulator, 0)


class Strchr(CStdModel):
    name = "strchr"

    # const char *strchr(const char *ptr, int value);
    argument_types = [ArgumentType.POINTER, ArgumentType.INT]
    return_type = ArgumentType.POINTER

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        ptr = self.get_arg1(emulator)
        val = self.get_arg2(emulator)

        assert isinstance(ptr, int)
        assert isinstance(val, int)

        n = _emu_strlen(emulator, ptr)

        data = emulator.read_memory(ptr, n)
        for i in range(0, n):
            if data[i] == val:
                self.set_return_value(emulator, ptr + i)
                return

        self.set_return_value(emulator, 0)


class Strcspn(CStdModel):
    name = "strcspn"

    # size_t strcspn(const char *str1, const char *str2);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.SIZE_T

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        # size_t strcspn(const char *str1, const char *str2);
        ptr1 = self.get_arg1(emulator)
        ptr2 = self.get_arg2(emulator)

        assert isinstance(ptr1, int)
        assert isinstance(ptr2, int)

        len1 = _emu_strlen(emulator, ptr1)
        len2 = _emu_strlen(emulator, ptr2)

        bytes1 = emulator.read_memory(ptr1, len1)
        bytes2 = emulator.read_memory(ptr2, len2)

        needles = {x for x in bytes2}

        for i in range(0, len1):
            if bytes1[i] in needles:
                self.set_return_value(emulator, i)
                return

        self.set_return_value(emulator, len1)


class Strpbrk(CStdModel):
    name = "strpbrk"

    # const char *strpbrk(const char *str1, const char *str2);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.POINTER

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        ptr1 = self.get_arg1(emulator)
        ptr2 = self.get_arg2(emulator)

        assert isinstance(ptr1, int)
        assert isinstance(ptr2, int)

        len1 = _emu_strlen(emulator, ptr1)
        len2 = _emu_strlen(emulator, ptr2)

        bytes1 = emulator.read_memory(ptr1, len1)
        bytes2 = emulator.read_memory(ptr2, len2)

        needles = {x for x in bytes2}

        for i in range(0, len1):
            if bytes1[i] in needles:
                self.set_return_value(emulator, ptr1 + i)
                return

        self.set_return_value(emulator, 0)


class Strrchr(CStdModel):
    name = "strrchr"

    # const char *strrchr(const char *ptr, int value);
    argument_types = [ArgumentType.POINTER, ArgumentType.INT]
    return_type = ArgumentType.POINTER

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        ptr = self.get_arg1(emulator)
        val = self.get_arg2(emulator)

        assert isinstance(ptr, int)
        assert isinstance(val, int)

        n = _emu_strlen(emulator, ptr)

        data = emulator.read_memory(ptr, n)

        for i in range(n - 1, -1, -1):
            if data[i] == val:
                self.set_return_value(emulator, ptr + i)
                return

        self.set_return_value(emulator, 0)


class Strspn(CStdModel):
    name = "strspn"

    # size_t strspn(const char *str1, const char *str2);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.SIZE_T

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        ptr1 = self.get_arg1(emulator)
        ptr2 = self.get_arg2(emulator)

        assert isinstance(ptr1, int)
        assert isinstance(ptr2, int)

        len1 = _emu_strlen(emulator, ptr1)
        len2 = _emu_strlen(emulator, ptr2)

        bytes1 = emulator.read_memory(ptr1, len1)
        bytes2 = emulator.read_memory(ptr2, len2)

        needles = {x for x in bytes2}

        for i in range(0, len1):
            if bytes1[i] not in needles:
                self.set_return_value(emulator, i)
                return

        self.set_return_value(emulator, len1)


class Strstr(CStdModel):
    name = "strstr"

    # const char *strstr(const char *str1, const char *str2);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.POINTER

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        ptr1 = self.get_arg1(emulator)
        ptr2 = self.get_arg2(emulator)

        assert isinstance(ptr1, int)
        assert isinstance(ptr2, int)

        len1 = _emu_strlen(emulator, ptr1)
        len2 = _emu_strlen(emulator, ptr2)

        bytes1 = emulator.read_memory(ptr1, len1)
        bytes2 = emulator.read_memory(ptr2, len2)

        for i in range(0, len1):
            if bytes1[i : i + len2] == bytes2:
                self.set_return_value(emulator, ptr1 + i)
                return

        self.set_return_value(emulator, 0)


class Strtok(CStdModel):
    name = "strtok"

    # char *strtok(char *str, const char *delimiters);
    argument_types = [ArgumentType.POINTER, ArgumentType.POINTER]
    return_type = ArgumentType.POINTER

    def __init__(self, address: int):
        super().__init__(address)
        self.ptr = 0

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        ptr1 = self.get_arg1(emulator)
        ptr2 = self.get_arg2(emulator)

        assert isinstance(ptr1, int)
        assert isinstance(ptr2, int)

        if ptr1 == 0:
            if self.ptr == 0:
                raise exceptions.EmulationError(
                    "strtok called with NULL when placeholder was NULL"
                )
            ptr1 = self.ptr

        len1 = _emu_strlen(emulator, ptr1)
        len2 = _emu_strlen(emulator, ptr2)

        bytes1 = emulator.read_memory(ptr1, len1)
        bytes2 = emulator.read_memory(ptr2, len2)

        if len1 == 0:
            # Empty string; we're out of tokens.
            self.set_return_value(emulator, 0)
            self.ptr = 0
            return

        # Non-empty string; we will have a token.
        self.set_return_value(emulator, ptr1)

        needles = {x for x in bytes2}
        for i in range(0, len1):
            if bytes1[i] in needles:
                emulator.write_memory(ptr1 + i, b"\0")
                self.ptr = ptr1 + i + 1
                return

        # Fall-through case; token lasts to end of string
        self.ptr = ptr1 + len1


class Memset(CStdModel):
    name = "memset"

    # void *memset(void *ptr, int value, size_t num);
    argument_types = [ArgumentType.POINTER, ArgumentType.INT, ArgumentType.SIZE_T]
    return_type = ArgumentType.POINTER

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        ptr = self.get_arg1(emulator)
        val = self.get_arg2(emulator)
        n = self.get_arg3(emulator)

        assert isinstance(ptr, int)
        assert isinstance(val, int)
        assert isinstance(n, int)

        data = bytes([val & 0xFF]) * n
        emulator.write_memory(ptr, data)

        self.set_return_value(emulator, ptr)


class Strerror(CStdModel):
    name = "strerror"

    # const char *strerror(int errno);
    argument_types = [ArgumentType.INT]
    return_type = ArgumentType.POINTER

    # Needs a string buffer for the description
    static_space_required = 64

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)

        errno = self.get_arg1(emulator)

        assert isinstance(errno, int)

        resolver = ErrnoResolver.for_platform(self.platform, self.abi)

        try:
            description = resolver.get_description(errno)
        except KeyError:
            description = f"Unknown error {errno}"

        descbytes = description.encode("utf-8") + b"\0"

        assert len(descbytes) <= self.static_space_required
        assert isinstance(self.static_buffer_address, int)
        emulator.write_memory(self.static_buffer_address, descbytes)
        self.set_return_value(emulator, self.static_buffer_address)


class Strlen(CStdModel):
    name = "strlen"

    # size_t strlen(const char *str);
    argument_types = [ArgumentType.POINTER]
    return_type = ArgumentType.SIZE_T

    def model(self, emulator: emulators.Emulator) -> None:
        super().model(emulator)
        ptr = self.get_arg1(emulator)

        assert isinstance(ptr, int)

        res = _emu_strlen(emulator, ptr)
        self.set_return_value(emulator, res)


__all__ = [
    "Memcpy",
    "Memmove",
    "Strcpy",
    "Strncpy",
    "Strcat",
    "Strncat",
    "Memcmp",
    "Strncmp",
    "Strcmp",
    "Strcoll",
    "Strxfrm",
    "Memchr",
    "Strchr",
    "Strcspn",
    "Strpbrk",
    "Strrchr",
    "Strspn",
    "Strstr",
    "Strtok",
    "Memset",
    "Strerror",
    "Strlen",
]



from ....posix.signal import (
    BsdSignal,
    Kill,
    Killpg,
    PthreadKill,
    PthreadSigmask,
    Sigaction,
    Sigaddset,
    Sigaltstack,
    Sigdelset,
    Sigemptyset,
    Sigfillset,
    Sighold,
    Sigignore,
    Siginterrupt,
    Sigismember,
    Sigpause,
    Sigpending,
    Sigprocmask,
    Sigqueue,
    Sigrelse,
    Sigset,
    Sigsuspend,
    Sigtimedwait,
    Sigwait,
    Sigwaitinfo,
)
from ..systemv import RiscV64SysVModel


class RiscV64SysVBsdSignal(BsdSignal, RiscV64SysVModel):
    pass


class RiscV64SysVKill(Kill, RiscV64SysVModel):
    pass


class RiscV64SysVKillpg(Killpg, RiscV64SysVModel):
    pass


class RiscV64SysVPthreadKill(PthreadKill, RiscV64SysVModel):
    pass


class RiscV64SysVPthreadSigmask(PthreadSigmask, RiscV64SysVModel):
    pass


class RiscV64SysVSigaction(Sigaction, RiscV64SysVModel):
    pass


class RiscV64SysVSigaddset(Sigaddset, RiscV64SysVModel):
    pass


class RiscV64SysVSigaltstack(Sigaltstack, RiscV64SysVModel):
    pass


class RiscV64SysVSigdelset(Sigdelset, RiscV64SysVModel):
    pass


class RiscV64SysVSigemptyset(Sigemptyset, RiscV64SysVModel):
    pass


class RiscV64SysVSigfillset(Sigfillset, RiscV64SysVModel):
    pass


class RiscV64SysVSighold(Sighold, RiscV64SysVModel):
    pass


class RiscV64SysVSigignore(Sigignore, RiscV64SysVModel):
    pass


class RiscV64SysVSiginterrupt(Siginterrupt, RiscV64SysVModel):
    pass


class RiscV64SysVSigismember(Sigismember, RiscV64SysVModel):
    pass


class RiscV64SysVSigpause(Sigpause, RiscV64SysVModel):
    pass


class RiscV64SysVSigpending(Sigpending, RiscV64SysVModel):
    pass


class RiscV64SysVSigprocmask(Sigprocmask, RiscV64SysVModel):
    pass


class RiscV64SysVSigqueue(Sigqueue, RiscV64SysVModel):
    pass


class RiscV64SysVSigrelse(Sigrelse, RiscV64SysVModel):
    pass


class RiscV64SysVSigset(Sigset, RiscV64SysVModel):
    pass


class RiscV64SysVSigsuspend(Sigsuspend, RiscV64SysVModel):
    pass


class RiscV64SysVSigtimedwait(Sigtimedwait, RiscV64SysVModel):
    pass


class RiscV64SysVSigwait(Sigwait, RiscV64SysVModel):
    pass


class RiscV64SysVSigwaitinfo(Sigwaitinfo, RiscV64SysVModel):
    pass


__all__ = [
    "RiscV64SysVBsdSignal",
    "RiscV64SysVKill",
    "RiscV64SysVKillpg",
    "RiscV64SysVPthreadKill",
    "RiscV64SysVPthreadSigmask",
    "RiscV64SysVSigaction",
    "RiscV64SysVSigaddset",
    "RiscV64SysVSigaltstack",
    "RiscV64SysVSigdelset",
    "RiscV64SysVSigemptyset",
    "RiscV64SysVSigfillset",
    "RiscV64SysVSighold",
    "RiscV64SysVSigignore",
    "RiscV64SysVSiginterrupt",
    "RiscV64SysVSigismember",
    "RiscV64SysVSigpause",
    "RiscV64SysVSigpending",
    "RiscV64SysVSigprocmask",
    "RiscV64SysVSigqueue",
    "RiscV64SysVSigrelse",
    "RiscV64SysVSigset",
    "RiscV64SysVSigsuspend",
    "RiscV64SysVSigtimedwait",
    "RiscV64SysVSigwait",
    "RiscV64SysVSigwaitinfo",
]



import typing
from dataclasses import dataclass, field


@dataclass(frozen=False)
class ElfSymbol:
    """ELF symbol struct

    Lief parses this data, but I can't keep the objects around.

    This adds one field to the typical ELF symbol, baseaddr.
    ELF symbol values are often relative,
    but relative to what is an annoying question:

    - In non-PIC binaries, the value is relative to zero.
    - In linked PIC binaries, the value is relative to the load address
    - In relocatables (.o and .ko), the value is relative to the section's address

    I'll figure out the base address; to get the absolute address,
    just add value to baseaddr.

    This also adds a data structure helper field, relas,
    so I can collect the relas linked to a symbol
    without a separate dict.
    """

    idx: int  # Symbol table index
    dynamic: bool  # Is this in the static or dynamic symbol table?
    name: str  # Symbol name
    type: int  # Symbol type
    bind: int  # Symbol binding
    visibility: int  # Symbol visibility
    shndx: int  # Symbol section index, or reserved flags
    defined: bool  # Is this symbol defined?
    value: int  # Symbol value
    size: int  # Symbol size
    baseaddr: int  # Base address for relative symbol values
    relas: typing.List["ElfRela"] = field(
        default_factory=list
    )  # List of associated relas


@dataclass(frozen=False)
class ElfRela:
    """ELF relocation struct

    Lief parses this data, but I can't keep the objects around.

    NOTE: offset is the absolute address this rela is modifying.
    Normally, it's optionally relative to the load address,
    or to the linked section's address in a .o or .ko.
    I don't want to deal with that.
    """

    offset: int  # Address of the relocation
    type: int  # Relocation type; platform-dependent
    symbol: ElfSymbol  # Relevant symbol in use
    addend: int  # Extra argument to computation



from ....c99 import (
    Memchr,
    Memcmp,
    Memcpy,
    Memmove,
    Memset,
    Strcat,
    Strchr,
    Strcmp,
    Strcoll,
    Strcpy,
    Strcspn,
    Strerror,
    Strlen,
    Strncat,
    Strncmp,
    Strncpy,
    Strpbrk,
    Strrchr,
    Strspn,
    Strstr,
    Strtok,
    Strxfrm,
)
from ..systemv import AMD64SysVModel


class AMD64SysVMemcpy(Memcpy, AMD64SysVModel):
    pass


class AMD64SysVMemmove(Memmove, AMD64SysVModel):
    pass


class AMD64SysVStrcpy(Strcpy, AMD64SysVModel):
    pass


class AMD64SysVStrncpy(Strncpy, AMD64SysVModel):
    pass


class AMD64SysVStrcat(Strcat, AMD64SysVModel):
    pass


class AMD64SysVStrncat(Strncat, AMD64SysVModel):
    pass


class AMD64SysVMemcmp(Memcmp, AMD64SysVModel):
    pass


class AMD64SysVStrncmp(Strncmp, AMD64SysVModel):
    pass


class AMD64SysVStrcmp(Strcmp, AMD64SysVModel):
    pass


class AMD64SysVStrcoll(Strcoll, AMD64SysVModel):
    pass


class AMD64SysVStrxfrm(Strxfrm, AMD64SysVModel):
    pass


class AMD64SysVMemchr(Memchr, AMD64SysVModel):
    pass


class AMD64SysVStrchr(Strchr, AMD64SysVModel):
    pass


class AMD64SysVStrcspn(Strcspn, AMD64SysVModel):
    pass


class AMD64SysVStrpbrk(Strpbrk, AMD64SysVModel):
    pass


class AMD64SysVStrrchr(Strrchr, AMD64SysVModel):
    pass


class AMD64SysVStrspn(Strspn, AMD64SysVModel):
    pass


class AMD64SysVStrstr(Strstr, AMD64SysVModel):
    pass


class AMD64SysVStrtok(Strtok, AMD64SysVModel):
    pass


class AMD64SysVMemset(Memset, AMD64SysVModel):
    pass


class AMD64SysVStrerror(Strerror, AMD64SysVModel):
    pass


class AMD64SysVStrlen(Strlen, AMD64SysVModel):
    pass


__all__ = [
    "AMD64SysVMemcpy",
    "AMD64SysVMemmove",
    "AMD64SysVStrcpy",
    "AMD64SysVStrncpy",
    "AMD64SysVStrcat",
    "AMD64SysVStrncat",
    "AMD64SysVMemcmp",
    "AMD64SysVStrncmp",
    "AMD64SysVStrcmp",
    "AMD64SysVStrcoll",
    "AMD64SysVStrxfrm",
    "AMD64SysVMemchr",
    "AMD64SysVStrchr",
    "AMD64SysVStrcspn",
    "AMD64SysVStrpbrk",
    "AMD64SysVStrrchr",
    "AMD64SysVStrspn",
    "AMD64SysVStrstr",
    "AMD64SysVStrtok",
    "AMD64SysVMemset",
    "AMD64SysVStrerror",
    "AMD64SysVStrlen",
]



import logging
import typing

import angr
import claripy

from ... import analyses, emulators, exceptions, hinting, platforms, state
from ...emulators.angr.exceptions import PathTerminationSignal
from .. import forced_exec, underlays
from .guards import GuardTrackingScratchPlugin
from .hints import (
    ClaripySerializable,
    FieldEventHint,
    PartialBitFieldAccessHint,
    PartialByteFieldAccessHint,
    PartialByteFieldWriteHint,
    TrackedFieldHint,
    UnknownFieldHint,
)

log = logging.getLogger(__name__)

# Tell angr to be quiet please
logging.getLogger("angr").setLevel(logging.WARNING)


class FDAState:
    def __init__(self):
        self.fda_labels = set()
        self.fda_addr_to_label = dict()
        self.fda_label_to_addr = dict()
        self.fda_mem_ranges = set()
        self.fda_bindings = dict()
        self.fda_byte_labels = set()
        self.fda_addr_to_unk_label = dict()


class FieldDetectionMixin(underlays.AnalysisUnderlay):
    """Analysis comparing state labels to field accesses.

    This assumes that memory labels correspond to fields
    of a structure, or a struct-coded buffer.

    The hypothesis is that any memory accesses that don't correspond
    to a defined field mean that your understanding of
    the as-implemented program is incorrect.

    This is an initial experiment,
    but the goal is to either help lift data formats,
    or compare a program against a known format.

    This approach works best against fixed-size binary data formats.
    It's got a limited ability to handle run-length-encoded formats
    by concretizing the run lengths.
    Attempting to handle a delimited or text-based format
    will likely lead to insanity.
    """

    name = "Field Detection Analysis"
    version = "0.0"
    description = "Detects discrepancies between labels and field accesses"

    halt_on_hint = True

    def translate_global_addr(self, fda, addr):
        # Convert address into a label
        for r, label in fda.fda_addr_to_label.items():
            if addr in r:
                return label
        return "UNKNOWN"

    def add_unk_label(self, emu, fda, addr, expr):
        size = len(expr) // 8
        r = range(addr, addr + size)
        label = f"UNK.{hex(addr)}"
        sym = claripy.BVS(f"UNK.{hex(addr)}", size * 8)
        emu.add_constraint(sym == expr)

        fda.fda_addr_to_unk_label[r] = label
        return sym

    def mem_read_hook(self, emu, addr, size, expr):
        fda = emu.get_extension("fda")
        cheat = self.translate_global_addr(fda, addr)
        log.warning(
            f"{hex(emu.state._ip.concrete_value)}: READING {hex(addr)} - {hex(addr + size)} ({cheat})"
        )
        log.warning(f"  {expr}")

        r = range(addr, addr + size)
        if r in fda.fda_addr_to_unk_label:
            log.warning("  Read from already-hinted field")
            return None

        if expr.op == "Reverse":
            # Bytes are reversed.
            # angr's memory is big-endian,
            # and the program interpreted it as a little-endian value.
            expr = expr.args[0]

        if expr.op == "BVV":
            # This is a concrete value.  These are safe
            return None

        if expr.op == "BVS":
            # This is a symbol.  It's okay if it's a field, or an UNK field
            label = expr.args[0].split("_")[0]
            if label in fda.fda_labels or label.startswith("UNK."):
                return

            hint = UnknownFieldHint(
                pc=emu.read_register("pc"),
                guards=list(
                    map(
                        lambda x: (x[0], ClaripySerializable(x[1])),
                        filter(lambda x: x[1].op != "BoolV", emu.state.scratch.guards),
                    )
                ),
                address=addr,
                size=size,
                expr=str(expr),
            )
            self.hinter.self(hint)
            if self.halt_on_hint:
                raise PathTerminationSignal()
            else:
                self.add_unk_label(emu, fda, addr, expr)
                return None

        if expr.op == "Extract":
            # This is a slice of another expression.
            if expr.args[2].op == "BVS":
                # This is a slice of a symbol, which means
                # the program read part of a field we labeled.
                # Dollars to ducats this means our labeling isn't precise enough.
                var = expr.args[2]
                label = var.args[0].split("_")[0]

                if var.args[0] not in fda.fda_labels:
                    # This is not a label we know
                    hint = UnknownFieldHint(
                        pc=emu.read_register("pc"),
                        guards=list(
                            map(
                                lambda x: (x[0], ClaripySerializable(x[1])),
                                filter(
                                    lambda x: x[1].op != "BoolV",
                                    emu.state.scratch.guards,
                                ),
                            )
                        ),
                        address=addr,
                        size=size,
                        access="read",
                        expr=str(var),
                    )
                else:
                    # This is a partial read from a known field
                    start = expr.args[0]
                    end = expr.args[1]

                    # Angr's bit numbering is annoying.
                    r = fda.fda_label_to_addr[var.args[0]]
                    field_size = r.stop - r.start
                    start = field_size * 8 - start - 1
                    end = field_size * 8 - end

                    if start % 8 != 0 or end % 8 != 0:
                        hint = PartialBitFieldAccessHint(
                            pc=emu.read_register("pc"),
                            guards=list(
                                map(
                                    lambda x: (x[0], ClaripySerializable(x[1])),
                                    filter(
                                        lambda x: x[1].op != "BoolV",
                                        emu.state.scratch.guards,
                                    ),
                                )
                            ),
                            address=addr,
                            size=size,
                            access="read",
                            label=var.args[0],
                            start=start,
                            end=end,
                        )
                    else:
                        hint = PartialByteFieldAccessHint(
                            pc=emu.read_register("pc"),
                            guards=list(
                                map(
                                    lambda x: (x[0], ClaripySerializable(x[1])),
                                    filter(
                                        lambda x: x[1].op != "BoolV",
                                        emu.state.scratch.guards,
                                    ),
                                )
                            ),
                            address=addr,
                            size=size,
                            access="read",
                            label=var.args[0],
                            start=start // 8,
                            end=end // 8,
                        )
                self.hinter.send(hint)
                if self.halt_on_hint:
                    raise PathTerminationSignal()
                else:
                    self.add_unk_label(emu, fda, addr, expr)
                    return None

        # This is a complex expression.
        # We're accessing the results of a computation
        hint = UnknownFieldHint(
            pc=emu.read_register("pc"),
            guards=list(
                map(
                    lambda x: (x[0], ClaripySerializable(x[1])),
                    filter(lambda x: x[1].op != "BoolV", emu.state.scratch.guards),
                )
            ),
            address=addr,
            size=size,
            access="read",
            expr=str(expr),
        )
        self.hinter.send(hint)
        if self.halt_on_hint:
            raise PathTerminationSignal()
        else:
            self.add_unk_label(emu, fda, addr, expr)
            return None

    def mem_write_hook(self, emu, addr, size, expr):
        fda = emu.get_extension("fda")
        cheat = self.translate_global_addr(fda, addr)
        log.warning(
            f"{hex(emu.state._ip.concrete_value)}: WRITING {hex(addr)} - {hex(addr + size)} ({cheat})"
        )
        log.warning(f"  {expr}")

        good = False
        bad = False

        r = range(addr, addr + size)
        if r in fda.fda_addr_to_unk_label:
            label = fda.fda_addr_to_unk_label[r]
            log.warning(f"  Write to already-hinted field {label}")
            sym = claripy.BVS(label, size * 8)
            emu.add_constraint(sym == expr)
            emu.state.inspect.mem_write_expr = sym

            return

        for r, label in fda.fda_addr_to_label.items():
            if addr in r:
                # Write is within an existing field
                if r.start == addr and r.stop == addr + size:
                    # Write lines up with the existing field.
                    # Create a new symbol for the new def of this field.
                    log.warning(f"  Write to entire field {cheat}")
                    sym = claripy.BVS(label, size * 8)
                    emu.add_constraint(sym == expr)
                    emu.state.inspect.mem_write_expr = sym
                    fda.fda_bindings[sym.args[0]] = expr
                    good = True
                else:
                    # Write does not line up with the existing field
                    start = addr - r.start
                    end = addr + size - r.start
                    hint = PartialByteFieldWriteHint(
                        pc=emu.read_register("pc"),
                        guards=list(
                            map(
                                lambda x: (x[0], ClaripySerializable(x[1])),
                                filter(
                                    lambda x: x[1].op != "BoolV",
                                    emu.state.scratch.guards,
                                ),
                            )
                        ),
                        address=addr,
                        size=size,
                        label=label,
                        start=start,
                        end=end,
                        expr=str(expr),
                    )
                    self.hinter.send(hint)
                    bad = True
        if bad and good:
            log.error("Write was complete and partial; your labels overlap.")
            raise exceptions.ConfigurationError("Overlapping labels")
        elif bad:
            if self.halt_on_hint:
                raise PathTerminationSignal()
            else:
                sym = self.add_unk_label(emu, fda, addr, expr)
                emu.state.inspect.mem_write_expr = sym
                return
        elif good:
            return
        else:
            # Write doesn't overlap any known fields
            hint = UnknownFieldHint(
                pc=emu.read_register("pc"),
                guards=list(
                    map(
                        lambda x: (x[0], ClaripySerializable(x[1])),
                        filter(lambda x: x[1].op != "BoolV", emu.state.scratch.guards),
                    )
                ),
                address=addr,
                size=size,
                access="write",
                expr=str(expr),
            )
            self.hinter.send(hint)
            if self.halt_on_hint:
                raise PathTerminationSignal()
            else:
                sym = self.add_unk_label(emu, fda, addr, expr)
                emu.state.inspect.mem_write_expr = sym
                return

    def angr_preinit(self, emu):
        preset = angr.SimState._presets["default"].copy()
        preset.add_default_plugin("scratch", GuardTrackingScratchPlugin)
        emu._plugin_preset = preset

    def run(self, machine):
        # Set up the filter analysis
        fda = FDAState()
        filt = FieldDetectionFilter()
        filt.activate()

        # Set up the emulator

        machine.apply(self.emulator)
        self.emulator.add_extension("fda", fda)

        # Capture the labeled memory ranges
        for s in machine:
            if isinstance(s, state.memory.Memory):
                start = s.address
                end = start + s.get_capacity()
                log.warning(f"{hex(start)} - {hex(end)}")
                for off, val in s.items():
                    label = val.get_label()
                    if label is not None:
                        # This is a labeled value.  It represents a field we want to track.
                        val_start = start + off
                        val_end = val_start + val.get_size()
                        log.warning(
                            f"  {hex(val_start)} - {hex(val_end)}: {val} := {label}"
                        )
                        hint = TrackedFieldHint(
                            message="Tracking new field",
                            address=val_start,
                            size=val.get_size(),
                            label=label,
                        )
                        self.hinter.send(hint)

                        if label in fda.fda_labels:
                            log.error(
                                f"You reused label {label}; please give it a unique name"
                            )
                            raise exceptions.ConfigurationError("Duplicate field name")

                        r = range(val_start, val_end)
                        fda.fda_labels.add(label)
                        fda.fda_addr_to_label[r] = label
                        fda.fda_label_to_addr[label] = r
                        fda.fda_bindings[label] = claripy.BVS(
                            label, val.get_size() * 8, explicit_name=True
                        )
                        fda.fda_mem_ranges.add((val_start, val_end))

        for start, end in fda.fda_mem_ranges:
            self.emulator.hook_memory_read_symbolic(start, end, self.mem_read_hook)
            self.emulator.hook_memory_write_symbolic(start, end, self.mem_write_hook)

        self.execute()

        log.warning(self.emulator.mgr)
        log.warning(self.emulator.mgr.errored)

        def state_visitor(emu: emulators.Emulator) -> None:
            if not isinstance(emu, emulators.AngrEmulator):
                raise TypeError(type(emu))
            fda = emu.get_extension("fda")
            for start, end in fda.fda_mem_ranges:
                emu.unhook_memory_read(start, end)
                emu.unhook_memory_write(start, end)

            log.warning(f"State at {emu.state._ip}:")
            log.warning("  Guards:")
            for ip, guard in emu.state.scratch.guards:
                if guard.op == "BoolV":
                    continue
                log.warning(f"    {hex(ip)}: {guard}")
            log.warning("  Fields:")
            for r in fda.fda_addr_to_label:
                val = emu.state.memory.load(r.start, r.stop - r.start)
                if len(val.variables) == 1:
                    # val is extremely likely a bound symbol.
                    # Fetch the binding
                    (label,) = val.variables
                    if label in fda.fda_bindings:
                        val = fda.fda_bindings[label]
                    else:
                        log.error(f"  Unknown variable {label}")

                log.warning(f"    {hex(r.start)} - {hex(r.stop)}: {label} = {val}")

        self.emulator.visit_states(state_visitor, stash="deadended")
        filt.deactivate()


class FieldDetectionFilter(analyses.Analysis):
    """Secondary field definition analysis.

    This picks up patterns that aren't noticeable from any
    single field access detection.
    """

    name = "field-detection-filter"
    version = "0.0"
    description = ""

    def __init__(self):
        super().__init__()
        self.active = True
        self.partial_ranges = dict()

    def analyze(self, hint: hinting.Hint):
        # Step 0: Print hints in a sane format.
        # The raw hint logging is unreadable.
        if not isinstance(hint, FieldEventHint):
            return
        hint.pp(log.error)

        if isinstance(hint, PartialByteFieldAccessHint):
            # Step 1: If this is a field access, remember it for later
            field = (hint.start, hint.end)
            self.partial_ranges.setdefault(hint.label, dict()).setdefault(
                field, list()
            ).append(hint)

    def activate(self):
        self.listen(FieldEventHint, self.analyze)

    def deactivate(self):
        super().deactivate()
        if not self.active:
            return
        self.active = False
        # Post-process results to detect trends we'd miss on single hints
        solver = claripy.solvers.Solver()

        for label, fields in self.partial_ranges.items():
            # Detect multiple fields at the same label/offset
            # Step 1: Collate the field info for this label
            fields_by_start = dict()
            for (start, end), hints in fields.items():
                guards = fields_by_start.setdefault(start, dict()).setdefault(
                    end, set()
                )
                for hint in hints:
                    # Unify all guards across all hints.
                    # I'm only going to care about pairwise matches.
                    guards.update(
                        map(lambda x: x.expr, map(lambda x: x[1], hint.guards))
                    )

            # Step 2: See if we actually have overlaps
            for start, rest in fields_by_start.items():
                if len(rest) > 1:
                    log.error(f"Multiple field defs starting at {label}[{start}]:")
                    for end, guards in rest.items():
                        log.error(f"    {label}[{start}:{end}]")

                    # Step 3: see if we can pinpoint the controlling variables
                    # The way angr works, two states will have compilmentary guards;
                    # guard vs !guard.  If we can find these expressions,
                    # we have a short list of known fields to consider
                    control_vars = set()
                    for a_end, a_guards in rest.items():
                        for b_end, b_guards in rest.items():
                            for a_guard in a_guards:
                                for b_guard in b_guards:
                                    # Forgive me, Prof. Sleator, for I have sinned...
                                    # I ran an SMT solver in a quadruple for-loop.
                                    if not solver.satisfiable([a_guard == b_guard]):
                                        control_vars.update(a_guard.variables)
                                        control_vars.update(b_guard.variables)
                    if len(control_vars) != 0:
                        log.error(
                            "  One or more of the following fields may control the format,"
                        )
                        log.error("  either as a flag, type code or length:")
                        for var in control_vars:
                            log.error(f"    {var}")


class FieldDetectionAnalysis(FieldDetectionMixin, underlays.BasicAnalysisUnderlay):
    """Detect fields on full path exploration"""

    name = "Field Detection Analysis"
    version = "0.0"
    description = "Detects discrepancies between labels and field accesses"

    def __init__(self, platform: platforms.Platform):
        self.platform = platform
        self.emulator = emulators.AngrEmulator(self.platform, preinit=self.angr_preinit)


class ForcedFieldDetectionAnalysis(
    FieldDetectionMixin, forced_exec.ForcedExecutionUnderlay
):
    name = "Field Detection Analysis - Forced"
    version = "0.0"
    description = "Detects discrepancies between labels and field accesses"

    halt_on_hint = False

    def __init__(
        self, platform: platforms.Platform, trace: typing.List[typing.Dict[str, int]]
    ):
        self.platform = platform
        self.emulator = emulators.AngrEmulator(self.platform, preinit=self.angr_preinit)
        self.emulator.enable_linear()
        super().__init__(trace)



import typing
from dataclasses import dataclass


@dataclass(frozen=False)
class PEExport:
    """PE export struct

    Lief parses this data, but I can't keep the objects around.

    An export defines a symbol that can be imported by other images.
    See 'PEImport' for more on that process.

    This is a mashup of a number of PE32+ structs
    (export definitions are actually somewhat complicated
    compared to import definitions).

    Exports can be referenced by name or by ordinal - a numeric key
    specific to the DLL.

    Exports can provide an actual address, or a "forwarder".
    This is the name of another export in another DLL that
    will provide the actual value of the symbol... or possibly forward again.
    """

    dll: str
    ordinal: int
    name: str
    forwarder: typing.Optional[str]
    value: int


@dataclass(frozen=False)
class PEImport:
    """PE import struct

    Lief parses this data, but I can't keep the objects around.

    An import defines that a specific exported symbol
    is required from a specific DLL file.

    They are closer to a relocation entry than a symbol;
    each import defines the export required, and where
    to write the exported value.  There is exactly one
    relocation mechanism, making this a lot easier to parse than ELF.

    This class actually mashes up two PE32+ structures:
    an Import Directory Entry, which defines a DLL that the file needs,
    and an Import Lookup Entry, which defines a specific export it needs.

    Lookup requires two keys, either the DLL name and export name,
    or the DLL and export ordinal - a numeric key specified wihtin the DLL.
    """

    dll: str  # Name of the DLL
    name: typing.Optional[str]  # String name of the export
    ordinal: typing.Optional[int]  # Ordinal of the export
    iat_address: int  # Address of the IAT slot for this import
    forwarder: typing.Optional[str]  # Forwarder export, if provided
    value: typing.Optional[int]  # Value imported



from __future__ import annotations

import logging
import sys
import typing
from enum import Enum

import capstone
import claripy
import unicorn
import unicorn.ppc_const  # Not properly exposed by the unicorn module

from ... import exceptions, instructions, platforms, utils
from .. import emulator, hookable
from .machdefs import UnicornMachineDef

logger = logging.getLogger(__name__)


class UnicornEmulationError(exceptions.EmulationError):
    def __init__(self, uc_err: unicorn.UcError, pc: int, msg: str, details: dict):
        self.uc_err = uc_err
        self.pc = pc
        self.msg = msg
        self.details = details

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self.uc_err}, {hex(self.pc)}, {self.details})"
        )


class UnicornEmulationMemoryFetchError(UnicornEmulationError):
    pass


class UnicornEmulationMemoryReadError(UnicornEmulationError):
    pass


class UnicornEmulationMemoryWriteError(UnicornEmulationError):
    pass


class UnicornEmulationExecutionError(UnicornEmulationError):
    pass


class EmulatorState(Enum):
    START_BLOCK = 1
    START_STEP = 2
    DELAY_STEP = 3
    STEP = 4
    BLOCK = 5
    RUN = 6
    SETUP = 7
    EXIT = 8


class UnicornEmulator(
    emulator.Emulator,
    hookable.QInstructionHookable,
    hookable.QFunctionHookable,
    hookable.QMemoryReadHookable,
    hookable.QMemoryWriteHookable,
    hookable.QInterruptHookable,
):
    """An emulator for the Unicorn emulation engine."""

    description = "This is a smallworld class encapsulating the Unicorn emulator."
    name = "smallworld's-unicorn"
    version = "0.0"

    PAGE_SIZE = 0x1000

    def __init__(self, platform: platforms.Platform):
        super().__init__(platform)
        self.platform = platform
        self.platdef = platforms.PlatformDef.for_platform(self.platform)
        self.machdef = UnicornMachineDef.for_platform(self.platform)
        self.engine = unicorn.Uc(self.machdef.uc_arch, self.machdef.uc_mode)
        self.disassembler = capstone.Cs(
            self.platdef.capstone_arch, self.platdef.capstone_mode
        )
        self.disassembler.detail = True

        self.memory_map: utils.RangeCollection = utils.RangeCollection()
        self.state: EmulatorState = EmulatorState.SETUP
        # labels are per byte

        # We'll have one entry in this dictionary per full-width base
        # register (by name) and those themselves are a map from offset
        # within the register to string label.
        # In other words, for 64-bit x86, we'd have
        # self.label["rax"][0] = "input", e.g., for the 0th byte in rax.
        # But we will not have self.label["eax"] -- you have to look in "rax"
        # Note that read_register_label will navigate that for you...
        # For memory, we have one label per byte in memory with address
        # translated via `hex(address)`.
        # In other words, self.label["0xdeadbeef"] = "came_from_hades" is
        # the label on that address in memory
        self.label: typing.Dict[str, typing.Dict[int, str]] = {}

        # this will run on *every instruction
        def code_callback(uc, address, size, user_data):
            # Check if we're out of bounds
            if not self._bounds.is_empty() and not self._bounds.contains_value(address):
                self.state = EmulatorState.EXIT
                self.engine.emu_stop()
                raise exceptions.EmulationBounds

            # check for if we've hit an exit point
            if address in self._exit_points:
                logger.debug(f"stopping emulation at exit point {address:x}")
                self.state = EmulatorState.EXIT
                self.engine.emu_stop()
                raise exceptions.EmulationExitpoint

            # Check single-step conditions.
            #
            # This callback gets invoked before the instruction is emulated.
            # When single-stepping, we want to run through it once,
            # and then stop emulation when it's run the second time.
            #
            # NOTE: Calling emu_stop() doesn't immediately stop emulation.
            # Returning here prevents spurious instruction and function hooks.
            # The memory read/write hooks also check for EXIT
            # to ensure they don't get called twice on the same event.
            if self.state == EmulatorState.STEP:
                self.state = EmulatorState.EXIT
                self.engine.emu_stop()
                return
            if self.state == EmulatorState.DELAY_STEP:
                self.state = EmulatorState.STEP
            if self.state == EmulatorState.START_STEP:
                insn = self.current_instruction()
                if insn.mnemonic in self.platdef.delay_slot_mnemonics:
                    self.state = EmulatorState.DELAY_STEP
                else:
                    self.state = EmulatorState.STEP

            # run instruciton hooks
            if self.all_instructions_hook:
                self.all_instructions_hook(self)

            if cb := self.is_instruction_hooked(address):
                logger.debug(f"hit hooking address for instruction at {address:x}")
                cb(self)
            # check function hooks *before* bounds since these might be out-of-bounds
            if cb := self.is_function_hooked(address):
                logger.debug(
                    f"hit hooking address for function at {address:x} -- {self.function_hooks[address]}"
                )
                # note that hooking a function means that we stop at function
                # entry and, after running the hook, we do not let the function
                # execute. Instead, we return from the function as if it ran.
                # this permits modeling
                # this is the model for the function
                cb(self)

                # Mimic a platform-specific "return" instruction.
                if self.platform.architecture == platforms.Architecture.X86_32:
                    # i386: pop a 4-byte value off the stack
                    sp = self.read_register("esp")
                    ret = int.from_bytes(
                        self.read_memory(sp, 4), self.platform.byteorder.value
                    )
                    self.write_register("esp", sp + 4)
                elif self.platform.architecture == platforms.Architecture.X86_64:
                    # amd64: pop an 8-byte value off the stack
                    sp = self.read_register("rsp")
                    ret = int.from_bytes(
                        self.read_memory(sp, 8), self.platform.byteorder.value
                    )
                    self.write_register("rsp", sp + 8)
                elif (
                    self.platform.architecture == platforms.Architecture.AARCH64
                    or self.platform.architecture == platforms.Architecture.ARM_V5T
                    or self.platform.architecture == platforms.Architecture.ARM_V6M
                    or self.platform.architecture
                    == platforms.Architecture.ARM_V6M_THUMB
                    or self.platform.architecture == platforms.Architecture.ARM_V7A
                    or self.platform.architecture == platforms.Architecture.ARM_V7M
                    or self.platform.architecture == platforms.Architecture.ARM_V7R
                    or self.platform.architecture == platforms.Architecture.POWERPC32
                    or self.platform.architecture == platforms.Architecture.POWERPC64
                ):
                    # aarch64, arm32, powerpc and powerpc64: branch to register 'lr'
                    ret = self.read_register("lr")
                elif (
                    self.platform.architecture == platforms.Architecture.MIPS32
                    or self.platform.architecture == platforms.Architecture.MIPS64
                ):
                    # mips32 and mips64: branch to register 'ra'
                    ret = self.read_register("ra")
                else:
                    raise exceptions.ConfigurationError(
                        "Don't know how to return for {self.platform.architecture}"
                    )

                self.write_register("pc", ret)

        self.engine.hook_add(unicorn.UC_HOOK_CODE, code_callback)

        # functions to run before memory read and write for
        # specific addresses

        def mem_read_callback(uc, type, address, size, value, user_data):
            assert type == unicorn.UC_MEM_READ

            if self.state == EmulatorState.EXIT:
                # Spurious call during the end of a single-step
                #
                # It looks like the emulator may continue processing an instruction
                # even after emu_stop() is called.
                return

            orig_data = value.to_bytes(size, self.platform.byteorder.value)
            if self.all_reads_hook:
                data = self.all_reads_hook(self, address, size, orig_data)
                if data:
                    if len(data) != size:
                        raise exceptions.EmulationError(
                            f"Read hook at {hex(address)} returned {len(data)} bytes; need {size} bytes"
                        )
                    uc.mem_write(address, data)
                    orig_data = data

            if cb := self.is_memory_read_hooked(address, size):
                data = cb(self, address, size, orig_data)

                # Execute registered callback
                # data = cb(self, address, size)
                # Overwrite memory being read.
                # The instruction is emulated after this callback fires,
                # so the new value will get used for computation.
                if data:
                    if len(data) != size:
                        raise exceptions.EmulationError(
                            f"Read hook at {hex(address)} returned {len(data)} bytes; need {size} bytes"
                        )
                    uc.mem_write(address, data)

        def mem_write_callback(uc, type, address, size, value, user_data):
            assert type == unicorn.UC_MEM_WRITE

            if self.state == EmulatorState.EXIT:
                # Spurious call during the end of a single-step
                #
                # It looks like the emulator may continue processing an instruction
                # even after emu_stop() is called.
                return

            if self.all_writes_hook:
                self.all_writes_hook(
                    self,
                    address,
                    size,
                    value.to_bytes(size, self.platform.byteorder.value),
                )

            if cb := self.is_memory_write_hooked(address, size):
                cb(
                    self,
                    address,
                    size,
                    value.to_bytes(size, self.platform.byteorder.value),
                )

        self.engine.hook_add(unicorn.UC_HOOK_MEM_WRITE, mem_write_callback)
        self.engine.hook_add(unicorn.UC_HOOK_MEM_READ, mem_read_callback)

        # function to run on *every* interrupt
        self.interrupts_hook: typing.Optional[
            typing.Callable[[emulator.Emulator, int], None]
        ] = None

        # function to run on a specific interrupt number
        self.interrupt_hook: typing.Dict[
            int, typing.Callable[[emulator.Emulator], None]
        ] = {}

        def interrupt_callback(uc, index, user_data):
            if self.interrupts_hook is not None:
                self.interrupts_hook()
            if index in self.interrupt_hook:
                self.interrupt_hook[index]()

        self.engine.hook_add(unicorn.UC_HOOK_INTR, interrupt_callback)

        def block_callback(uc, address, block_size, user_data):
            if self.state == EmulatorState.BLOCK:
                if address != self.read_register("pc"):
                    # If a THUMB block jumps to an ARM block,
                    # the reported PC will be the address of the jump,
                    # not the address of the current block.
                    # The rest of the registers behave as if the jump had happened.
                    # On arm32, this includes flipping execution modes on a `blx` instruction.
                    #
                    # Manually updating PC from the address supplied
                    # in this callback fixes the problem.
                    self.write_register("pc", address)
                self.state = EmulatorState.EXIT
                self.engine.emu_stop()
            if self.state == EmulatorState.START_BLOCK:
                self.state = EmulatorState.BLOCK

        self.engine.hook_add(unicorn.UC_HOOK_BLOCK, block_callback)

        # keep track of which registers have been initialized
        self.initialized_registers: typing.Dict[str, typing.Set[int]] = {}

    def _check_pc_ok(self, pc):
        """Check if this pc is ok to emulate, i.e. in bounds and not an exit
        point."""

        if not self._bounds.is_empty() and not self._bounds.contains_value(pc):
            # There are bounds, and we are not in them
            return False

        # check for if we've hit an exit point
        if pc in self._exit_points:
            logger.debug(f"stopping emulation at exit point {pc:x}")
            return False
        return True

    def _register(self, name: str) -> typing.Tuple[typing.Any, str, int, int]:
        # Translate register name into the tuple
        # (u, b, o, s)
        # u is the unicorn reg number
        # b is the name of full-width base register this is or is part of
        # o is start offset within full-width base register
        # s is size in bytes
        name = name.lower()
        # support some generic register references
        if name == "pc":
            name = self.platdef.pc_register

        uc_const = self.machdef.uc_reg(name)
        reg = self.platdef.registers[name]

        if hasattr(reg, "parent"):
            parent = reg.parent
            offset = reg.offset
        else:
            parent = reg.name
            offset = 0

        return (uc_const, parent, reg.size, offset)

    def read_register_content(self, name: str) -> int:
        (reg, _, _, _) = self._register(name)
        if reg == 0:
            raise exceptions.UnsupportedRegisterError(
                "Unicorn does not support register {name} for {self.platform}"
            )
        try:
            return self.engine.reg_read(reg)
        except Exception as e:
            raise exceptions.AnalysisError(f"Failed reading {name} (id: {reg})") from e

    def read_register_label(self, name: str) -> typing.Optional[str]:
        (_, base_reg, size, offset) = self._register(name)
        if base_reg in self.label:
            # we'll return a string repr of set of labels on all byte offsets
            # for this register
            labels = set([])
            for i in range(offset, offset + size):
                if i in self.label[base_reg]:
                    label = self.label[base_reg][i]
                    if label is not None:
                        labels.add(label)
            return ":".join(list(labels))
        return None

    def read_register(self, name: str) -> int:
        return self.read_register_content(name)

    def write_register_content(
        self, name: str, content: typing.Union[None, int, claripy.ast.bv.BV]
    ) -> None:
        if content is None:
            logger.debug(f"ignoring register write to {name} - no value")
            return

        if isinstance(content, claripy.ast.bv.BV):
            raise exceptions.SymbolicValueError(
                "This emulator cannot handle bitvector expressions"
            )

        if name == "pc":
            content = self._handle_thumb_interwork(content)

        (reg, base_reg, size, start_offset) = self._register(name)
        try:
            self.engine.reg_write(reg, content)
        except Exception as e:
            raise exceptions.AnalysisError(f"Failed writing {name} (id: {reg})") from e
        # keep track of which bytes in this register have been initialized
        if base_reg not in self.initialized_registers:
            self.initialized_registers[base_reg] = set([])
        for o in range(start_offset, start_offset + size):
            self.initialized_registers[base_reg].add(o)
        logger.debug(f"set register {name}={content}")

    def write_register_label(
        self, name: str, label: typing.Optional[str] = None
    ) -> None:
        if label is None:
            return
        (_, base_reg, size, offset) = self._register(name)
        if base_reg not in self.label:
            self.label[base_reg] = {}
        for i in range(offset, offset + size):
            self.label[base_reg][i] = label

    def write_register(
        self, name: str, content: typing.Union[None, int, claripy.ast.bv.BV]
    ) -> None:
        self.write_register_content(name, content)

    def read_memory_content(self, address: int, size: int) -> bytes:
        if size > sys.maxsize:
            raise ValueError(f"{size} is too large (max: {sys.maxsize})")
        try:
            return bytes(self.engine.mem_read(address, size))
        except unicorn.UcError as e:
            logger.warn(f"Unicorn raised an exception on memory read {e}")
            self._error(e, "mem")
            assert False  # Line is unreachable

    def read_memory_label(self, address: int, size: int) -> typing.Optional[str]:
        labels = set()
        if "mem" not in self.label:
            return None
        else:
            for a in range(address, address + size):
                if a in self.label["mem"]:
                    labels.add(self.label["mem"][a])
            if len(labels) == 0:
                return None
            return ":".join(list(labels))

    def read_memory(self, address: int, size: int) -> bytes:
        return self.read_memory_content(address, size)

    def map_memory(self, address: int, size: int) -> None:
        # Round address down to a page boundary
        page_address = (address // self.PAGE_SIZE) * self.PAGE_SIZE

        # Expand the size to accound for moving address
        page_size = size + address - page_address

        # Round page_size up to the next page
        page_size = (
            (page_size + self.PAGE_SIZE - 1) // self.PAGE_SIZE
        ) * self.PAGE_SIZE

        # Fill in any gaps in the specified region
        region = (page_address, page_address + page_size)
        missing_ranges = self.memory_map.get_missing_ranges(region)

        for start, end in missing_ranges:
            self.memory_map.add_range((start, end))
            self.engine.mem_map(start, end - start)

    def get_memory_map(self) -> typing.List[typing.Tuple[int, int]]:
        return list(self.memory_map.ranges)

    def _is_address_mapped(self, address):
        (ind, found) = self.memory_map.find_closest_range(address)
        return found

    def _is_address_range_mapped(self, address_range):
        (a, b) = address_range
        for address in range(a, b):
            if self._is_address_mapped(address) is False:
                return False
        return True

    def write_memory_content(
        self, address: int, content: typing.Union[bytes, claripy.ast.bv.BV]
    ) -> None:
        if content is None:
            raise ValueError(f"{self.__class__.__name__} requires concrete state")

        if isinstance(content, claripy.ast.bv.BV):
            raise exceptions.SymbolicValueError(
                "This emulator cannot handle bitvector expressions"
            )

        if len(content) > sys.maxsize:
            raise ValueError(f"{len(content)} is too large (max: {sys.maxsize})")

        if not len(content):
            raise ValueError("memory write cannot be empty")

        try:
            # print(f"write_memory: {content}")
            self.engine.mem_write(address, content)
        except unicorn.UcError as e:
            logger.warn(f"Unicorn raised an exception on memory write {e}")
            self._error(e, "mem")

        logger.debug(f"wrote {len(content)} bytes to 0x{address:x}")

    def write_memory_label(
        self, address: int, size: int, label: typing.Optional[str] = None
    ) -> None:
        if label is None:
            return
        if "mem" not in self.label:
            self.label["mem"] = dict()
        for a in range(address, address + size):
            self.label["mem"][a] = label

    def write_memory(
        self, address: int, content: typing.Union[bytes, claripy.ast.bv.BV]
    ) -> None:
        self.write_memory_content(address, content)

    def hook_instruction(
        self, address: int, function: typing.Callable[[emulator.Emulator], None]
    ) -> None:
        super(UnicornEmulator, self).hook_instruction(address, function)
        self.map_memory(address, self.PAGE_SIZE)

    def hook_function(
        self, address: int, function: typing.Callable[[emulator.Emulator], None]
    ) -> None:
        super(UnicornEmulator, self).hook_function(address, function)
        self.map_memory(address, self.PAGE_SIZE)

    def _disassemble(
        self, code: bytes, base: int, count: typing.Optional[int] = None
    ) -> typing.Tuple[typing.List[capstone.CsInsn], str]:
        instructions = self.disassembler.disasm(code, base)
        disassembly = []
        insns = []
        for i, instruction in enumerate(instructions):
            if count is not None and i >= count:
                break
            insns.append(instruction)
            disassembly.append(f"{instruction.mnemonic} {instruction.op_str}")
        return (insns, "\n".join(disassembly))

    def current_instruction(self) -> capstone.CsInsn:
        pc = self.read_register("pc")
        code = self.read_memory(pc, 15)
        if code is None:
            raise AssertionError("invalid state")
        for i in self.disassembler.disasm(code, pc):
            return i

    def _check(self) -> None:
        # check if it's ok to begin emulating
        # 1. pc must be set in order to emulate
        (_, base_name, size, offset) = self._register("pc")
        if (
            base_name in self.initialized_registers
            and len(self.initialized_registers[base_name]) == size
        ):
            # pc is fully initialized
            pass
        else:
            raise exceptions.ConfigurationError(
                "pc not initialized, emulation cannot start"
            )
        # 2. an exit point is also required
        if len(self._exit_points) == 0:
            raise exceptions.ConfigurationError(
                "at least one exit point must be set, emulation cannot start"
            )

    def _check_arm32_platform(self):
        """Check for ARM32 platform architecture"""
        return self.platform.architecture in [
            platforms.Architecture.ARM_V5T,
            platforms.Architecture.ARM_V6M,
            platforms.Architecture.ARM_V7A,
            platforms.Architecture.ARM_V7M,
            platforms.Architecture.ARM_V7R,
        ]

    def get_thumb(self) -> bool:
        if not self._check_arm32_platform():
            raise exceptions.ConfigurationError(
                "called get_thumb() on non-ARM32 system"
            )

        CPSR_THUMB_MODE_MASK = 0x20
        cpsr = self.engine.reg_read(unicorn.arm_const.UC_ARM_REG_CPSR)
        if cpsr & CPSR_THUMB_MODE_MASK:
            return True
        else:
            return False

    def set_thumb(self, enabled=True) -> None:
        if not self._check_arm32_platform():
            raise exceptions.ConfigurationError(
                "called set_thumb() on non-ARM32 system"
            )

        CPSR_THUMB_MODE_MASK = 0x20
        cpsr = self.engine.reg_read(unicorn.arm_const.UC_ARM_REG_CPSR)

        if enabled:
            self.engine.reg_write(
                unicorn.arm_const.UC_ARM_REG_CPSR, cpsr | CPSR_THUMB_MODE_MASK
            )
        elif cpsr & CPSR_THUMB_MODE_MASK:
            self.engine.reg_write(
                unicorn.arm_const.UC_ARM_REG_CPSR, cpsr ^ CPSR_THUMB_MODE_MASK
            )

    # Handle Thumb ISA exchange for ARM32
    def _handle_thumb_interwork(self, pc) -> int:
        if not self._check_arm32_platform():
            return pc

        # emu_start clears thumb mode if the low bit of pc != 1.
        # We use CPSR to determine if unicorn was previously in thumb
        # mode and set the low bit to 1 to maintain it. We also set
        # the mode of the disassembler.
        if self.get_thumb() or pc & 1:
            pc |= 1
            self.disassembler.mode = capstone.CS_MODE_THUMB
        else:
            self.disassembler.mode = capstone.CS_MODE_ARM

        return pc

    def step_instruction(self) -> None:
        self._check()
        self.state = EmulatorState.START_STEP

        pc = self.read_register("pc")
        if pc in self._exit_points:
            raise exceptions.EmulationExitpoint

        pc = self._handle_thumb_interwork(pc)

        if pc not in self.function_hooks:
            disas = self.current_instruction()
            logger.debug(f"single step at 0x{disas.address:x}: {disas}")

        try:
            self.engine.emu_start(pc, 0x0)

        except unicorn.UcError as e:
            if (
                e.errno == unicorn.UC_ERR_FETCH_UNMAPPED
                and self.read_register("pc") in self.function_hooks
            ):
                # probably we tried to execute call to code that's not mapped?
                pass
            else:
                # translate this unicorn error into something richer
                self._error(e, "exec")

    def step_block(self) -> None:
        self._check()
        pc = self.read_register("pc")
        pc = self._handle_thumb_interwork(pc)
        if pc in self._exit_points:
            raise exceptions.EmulationExitpoint

        disas = self.current_instruction()
        logger.info(f"step block at 0x{disas.address:x}: {disas}")
        try:
            self.state = EmulatorState.START_BLOCK
            self.engine.emu_start(pc, 0x0)

            self.state = EmulatorState.BLOCK
            pc = self.read_register("pc")
            pc = self._handle_thumb_interwork(pc)
            self.engine.emu_start(pc, 0x0)
        except unicorn.UcError as e:
            self._error(e, "exec")

    def run(self) -> None:
        self._check()
        self.state = EmulatorState.RUN

        logger.info(
            f"starting emulation at 0x{self.read_register('pc'):x}"
        )  # until 0x{self._exit_point:x}")

        try:
            pc = self.read_register("pc")
            pc = self._handle_thumb_interwork(pc)
            self.engine.emu_start(pc, 0x0)
        except exceptions.EmulationStop:
            pass
        except unicorn.UcError as e:
            self._error(e, "exec")

        logger.info("emulation complete")

    def _error(
        self, error: unicorn.UcError, typ: str
    ) -> typing.Dict[typing.Union[str, int], typing.Union[str, int, bytes]]:
        """Raises new exception from unicorn exception with extra details.

        Should only be run while single stepping.

        Arguments:
            error: Unicorn exception.

        Raises:
            UnicornEmulationError with extra details about the error.
        """

        pc = self.read_register("pc")

        try:
            # NB: can't use self.read_memory here since if it has an exception it will call _error, itself.
            code = bytes(self.engine.mem_read(pc, 16))
            # on arm32, update disassembler for ARM vs Thumb
            _ = self._handle_thumb_interwork(pc)
            insns, _ = self._disassemble(code, pc, 1)
            i = instructions.Instruction.from_capstone(insns[0])
        except:
            # looks like that code is not available
            i = None

        exc: typing.Type[exceptions.EmulationError] = exceptions.EmulationError

        if typ == "mem":
            prefix = "Failed memory access"
            exc = UnicornEmulationMemoryReadError
        if typ == "exec":
            prefix = "Quit emulation"
            exc = UnicornEmulationExecutionError
        else:
            prefix = "Unexpected Unicorn error"

        # rws is list of either reads or writes. get list of these
        # reads or writes that is not actually available, i.e. memory
        # not mapped
        def get_unavailable_rw(rws):
            out = []
            for rw in rws:
                if type(rw) is instructions.BSIDMemoryReferenceOperand:
                    a = rw.address(self)
                    if not (self._is_address_mapped(a)):
                        out.append((rw, a))
            return out

        details: typing.Dict[typing.Union[str, int], typing.Union[str, int, bytes]] = {}

        def details_str(details):
            for k, v in details.items():
                if k == "pc":
                    return f" pc=0x{pc:x}"
                if "_reads" in k or "_writes" in k:
                    s = ""
                    for rw, a in v:
                        rws = str(rw)
                        import re

                        foo = re.search(r"\((.*)\)$", rws)
                        rw = foo.groups()[0]
                        x = f"address=0x{a:x} i.e. [{rw}]"
                        if s == "":
                            s = x
                        else:
                            s = s + ", " + x
                    return s
            return "None"

        if error.errno == unicorn.UC_ERR_READ_UNMAPPED:
            msg = f"{prefix} due to read of unmapped memory"
            # actually this is a memory read error
            exc = UnicornEmulationMemoryReadError
            if i is not None:
                details["unmapped_reads"] = get_unavailable_rw(i.reads)
        elif error.errno == unicorn.UC_ERR_READ_PROT:
            msg = f"{prefix} due to read of mapped but protected memory"
            # actually this is a memory read error
            exc = UnicornEmulationMemoryReadError
            if i is not None:
                details["protected_reads"] = get_unavailable_rw(i.reads)
        elif error.errno == unicorn.UC_ERR_READ_UNALIGNED:
            msg = f"{prefix} due to unaligned read"
            # actually this is a memory read error
            exc = UnicornEmulationMemoryReadError
            if i is not None:
                details["unaligned_reads"] = get_unavailable_rw(i.reads)

        elif error.errno == unicorn.UC_ERR_WRITE_UNMAPPED:
            msg = f"{prefix} due to write to unmapped memory"
            # actually this is a memory write error
            exc = UnicornEmulationMemoryWriteError
            if i is not None:
                details["unmapped_writes"] = get_unavailable_rw(i.writes)
        elif error.errno == unicorn.UC_ERR_WRITE_PROT:
            msg = f"{prefix} due to write to mapped but protected memory"
            # actually this is a memory write error
            exc = UnicornEmulationMemoryWriteError
            if i is not None:
                details["protected_writes"] = get_unavailable_rw(i.writes)
        elif error.errno == unicorn.UC_ERR_WRITE_UNALIGNED:
            msg = f"{prefix} due to unaligned write"
            # actually this is a memory write error
            exc = UnicornEmulationMemoryWriteError
            if i is not None:
                details["unaligned_writes"] = get_unavailable_rw(i.writes)

        elif error.errno == unicorn.UC_ERR_FETCH_UNMAPPED:
            msg = f"{prefix} due to fetch of unmapped memory at"
            exc = UnicornEmulationMemoryFetchError
            details["pc"] = pc
            if pc in self._exit_points:
                # This is actually an exit point
                return details
            if not self._bounds.is_empty() and not self._bounds.contains_value(pc):
                # This is actually an out-of-bounds error
                return details
        elif error.errno == unicorn.UC_ERR_FETCH_PROT:
            msg = f"{prefix} due to fetch of from mapped but protected memory"
            exc = UnicornEmulationMemoryFetchError
            details["pc"] = pc
        elif error.errno == unicorn.UC_ERR_FETCH_UNALIGNED:
            msg = f"{prefix} due to unaligned fetch"
            exc = UnicornEmulationMemoryFetchError
            details["pc"] = pc

        elif error.errno == unicorn.UC_ERR_NOMEM:
            msg = f"{prefix} due Out-Of-Memory"
        elif error.errno == unicorn.UC_ERR_INSN_INVALID:
            msg = f"{prefix} due invalid instruction"
            details = {"pc": pc, "instr": str(i)}
        elif error.errno == unicorn.UC_ERR_RESOURCE:
            msg = f"{prefix} due insufficient resources"
        elif error.errno == unicorn.UC_ERR_EXCEPTION:
            msg = f"{prefix} due cpu exception"
        else:
            msg = f"{prefix} due to unknown Unicorn error {error.errno}"

        logger.warn(f"emulation stopped - reason: {error}")

        if i is None:
            logger.warn(
                f"FYI Unicorn rich exception processing unable to read code at pc=0x{pc:x} bc it is unavailable"
            )
        raise exc(error, pc, msg + " " + details_str(details), details)

    def __repr__(self) -> str:
        return f"UnicornEmulator(platform={self.platform})"


__all__ = [
    "UnicornEmulator",
    "UnicornEmulationMemoryReadError",
    "UnicornEmulationMemoryWriteError",
    "UnicornEmulationExecutionError",
]



from ....posix.libgen import Basename, Dirname
from ..systemv import ArmELSysVModel


class ArmELSysVBasename(Basename, ArmELSysVModel):
    pass


class ArmELSysVDirname(Dirname, ArmELSysVModel):
    pass


__all__ = [
    "ArmELSysVBasename",
    "ArmELSysVDirname",
]



from ....c99 import (
    Clearerr,
    Fclose,
    Feof,
    Ferror,
    Fflush,
    Fgetc,
    Fgetpos,
    Fgets,
    Fopen,
    Fprintf,
    Fputc,
    Fputs,
    Fread,
    Freopen,
    Fscanf,
    Fseek,
    Fsetpos,
    Ftell,
    Fwrite,
    Getc,
    Getchar,
    Gets,
    Printf,
    Putc,
    Putchar,
    Puts,
    Remove,
    Rename,
    Rewind,
    Scanf,
    Snprintf,
    Sprintf,
    Sscanf,
    Tmpfile,
    Tmpnam,
    Ungetc,
    Vfprintf,
    Vfscanf,
    Vprintf,
    Vscanf,
    Vsnprintf,
    Vsprintf,
    Vsscanf,
)
from ..systemv import I386SysVModel


class I386SysVFclose(Fclose, I386SysVModel):
    pass


class I386SysVFeof(Feof, I386SysVModel):
    pass


class I386SysVFerror(Ferror, I386SysVModel):
    pass


class I386SysVFgetc(Fgetc, I386SysVModel):
    pass


class I386SysVFgets(Fgets, I386SysVModel):
    pass


class I386SysVFopen(Fopen, I386SysVModel):
    pass


class I386SysVFprintf(Fprintf, I386SysVModel):
    pass


class I386SysVFputc(Fputc, I386SysVModel):
    pass


class I386SysVFputs(Fputs, I386SysVModel):
    pass


class I386SysVFread(Fread, I386SysVModel):
    pass


class I386SysVFscanf(Fscanf, I386SysVModel):
    pass


class I386SysVFseek(Fseek, I386SysVModel):
    pass


class I386SysVFtell(Ftell, I386SysVModel):
    pass


class I386SysVFwrite(Fwrite, I386SysVModel):
    pass


class I386SysVGetc(Getc, I386SysVModel):
    pass


class I386SysVGetchar(Getchar, I386SysVModel):
    pass


class I386SysVGets(Gets, I386SysVModel):
    pass


class I386SysVPrintf(Printf, I386SysVModel):
    pass


class I386SysVPutc(Putc, I386SysVModel):
    pass


class I386SysVPutchar(Putchar, I386SysVModel):
    pass


class I386SysVPuts(Puts, I386SysVModel):
    pass


class I386SysVRemove(Remove, I386SysVModel):
    pass


class I386SysVRename(Rename, I386SysVModel):
    pass


class I386SysVRewind(Rewind, I386SysVModel):
    pass


class I386SysVScanf(Scanf, I386SysVModel):
    pass


class I386SysVSnprintf(Snprintf, I386SysVModel):
    pass


class I386SysVSprintf(Sprintf, I386SysVModel):
    pass


class I386SysVSscanf(Sscanf, I386SysVModel):
    pass


class I386SysVClearerr(Clearerr, I386SysVModel):
    pass


class I386SysVFflush(Fflush, I386SysVModel):
    pass


class I386SysVFreopen(Freopen, I386SysVModel):
    pass


class I386SysVFgetpos(Fgetpos, I386SysVModel):
    pass


class I386SysVFsetpos(Fsetpos, I386SysVModel):
    pass


class I386SysVTmpfile(Tmpfile, I386SysVModel):
    pass


class I386SysVTmpnam(Tmpnam, I386SysVModel):
    pass


class I386SysVUngetc(Ungetc, I386SysVModel):
    pass


class I386SysVVfprintf(Vfprintf, I386SysVModel):
    pass


class I386SysVVfscanf(Vfscanf, I386SysVModel):
    pass


class I386SysVVprintf(Vprintf, I386SysVModel):
    pass


class I386SysVVscanf(Vscanf, I386SysVModel):
    pass


class I386SysVVsnprintf(Vsnprintf, I386SysVModel):
    pass


class I386SysVVsprintf(Vsprintf, I386SysVModel):
    pass


class I386SysVVsscanf(Vsscanf, I386SysVModel):
    pass


__all__ = [
    "I386SysVFclose",
    "I386SysVFeof",
    "I386SysVFerror",
    "I386SysVFgetc",
    "I386SysVFgets",
    "I386SysVFopen",
    "I386SysVFprintf",
    "I386SysVFputc",
    "I386SysVFputs",
    "I386SysVFread",
    "I386SysVFscanf",
    "I386SysVFseek",
    "I386SysVFtell",
    "I386SysVFwrite",
    "I386SysVGetc",
    "I386SysVGetchar",
    "I386SysVGets",
    "I386SysVPrintf",
    "I386SysVPutc",
    "I386SysVPutchar",
    "I386SysVPuts",
    "I386SysVRemove",
    "I386SysVRename",
    "I386SysVRewind",
    "I386SysVScanf",
    "I386SysVSnprintf",
    "I386SysVSprintf",
    "I386SysVSscanf",
    "I386SysVClearerr",
    "I386SysVFflush",
    "I386SysVFreopen",
    "I386SysVFgetpos",
    "I386SysVFsetpos",
    "I386SysVTmpfile",
    "I386SysVTmpnam",
    "I386SysVUngetc",
    "I386SysVVfprintf",
    "I386SysVVfscanf",
    "I386SysVVprintf",
    "I386SysVVscanf",
    "I386SysVVsnprintf",
    "I386SysVVsprintf",
    "I386SysVVsscanf",
]



import capstone

from ..platforms import Architecture, Byteorder
from .platformdef import PlatformDef, RegisterAliasDef, RegisterDef


class AMD64BasePlatformDef(PlatformDef):
    # Abstract platform def for AMD64
    #
    # AVX512 changes how the xmm and ymm registers are aliased.
    # The new bases are incompatible, so we need two separate classes.
    byteorder = Byteorder.LITTLE

    address_size = 8

    capstone_arch = capstone.CS_ARCH_X86
    capstone_mode = capstone.CS_MODE_64

    conditional_branch_mnemonics = {
        # All 57 varieties of Jcc opcode
        "jo",
        "jno",
        "jb",
        "jnae",
        "jc",
        "jnb",
        "jae",
        "jnc",
        "jz",
        "je",
        "jnz",
        "jne",
        "jbe",
        "jna",
        "jnbe",
        "ja",
        "js",
        "jns",
        "jp",
        "jpe",
        "jnp",
        "jpo",
        "jl",
        "jnge",
        "jnl",
        "jge",
        "jle",
        "jng",
        "jnle",
        "jg",
        # Jump if (re)?cx is zero
        # Oddly I can't find this in the opcode tables;
        # it's mentioned tangentially in the docs for Jcc
        # (and it's accepted by the assembler.)
        "jrcxz",
        "jecxz",
        "jcxz",
    }
    # TODO: Should arithmetic operations that impact flags be compares?
    compare_mnemonics = {
        # Basic integer comparisons.
        "cmp",
        "test",
        # SIMD float comparisons
        # 'pd' and 'ps' variants work on packed double and packed single floats.
        # TODO I'm not sure how this extends for AVX2 registers
        # NOTE: There's another, less-descriptive way to decode these
        # what does capstone use?
        "cmpeqpd",
        "cmpeqps",
        "cmpltpd",
        "cmpltps",
        "cmplepd",
        "cmpleps",
        "cmpunordpd",
        "cmpunordps",
        "cmpneqpd",
        "cmpneqps",
        "cmpnltpd",
        "cmpnltps",
        "cmpnlepd",
        "cmpnleps",
        "cmpordpd",
        "cmpordps",
        # Compare string operands
        "cmps",
        "cmpsw",
        "cmpsd",
        "cmpsq",
        # Single float comparisons
        # 'sd' and 'ss' variants work on double and single-precision floats.
        # TODO: I'm not sure how this extends for AVX2 registers
        "cmpeqsd",
        "cmpeqss",
        "cmpltsd",
        "cmpltss",
        "cmplesd",
        "cmpless",
        "cmpunordsd",
        "cmpunordss",
        "cmpneqsd",
        "cmpneqss",
        "cmpnltsd",
        "cmpnltss",
        "cmpnlesd",
        "cmpnless",
        "cmpordsd",
        "cmpordss",
        # TODO: Do cmpxchg, cmpxchg8b, cmpxchg16b count?
        # Generic float comparison
        # Unlike previous lists, sets multiple flags depending.
        "comisd",
        "comiss",
        # x87 float comparison
        # Sets FPU condition flags
        "fcom",
        "fcomp",
        "fcompp",
        # x87 float comparison
        # Sets eflags
        "fcomi",
        "fcomip",
        "fcomipp",
        # x87 integer comparison
        "ficom",
        "ficomp",
        # x87 test (comparison against zero)
        # Sets FPU condition flags
        "ftst",
        # x87 float comparison
        # Sets FPU condition flags
        # Behaves differently with NaNs
        "fucom",
        "fucomp",
        "fucompp",
        # x87 examine float
        # Sets FPU flags according to which type of float this is (normal, zero, NaN, infty...)
        "fxam",
        # TODO: There may be more of these... waiting...
    }

    pc_register = "rip"
    sp_register = "rsp"

    # NOTE: rbp and rsp are not considered general.
    # You _can_ use them as GPRs; it's very rarely a good idea.
    general_purpose_registers = [
        "rax",
        "rbx",
        "rcx",
        "rdx",
        "rdi",
        "rsi",
        "r8",
        "r9",
        "r10",
        "r11",
        "r12",
        "r13",
        "r14",
        "r15",
    ]
    registers = {
        # *** General Purpose Registers ***
        "rax": RegisterDef(name="rax", size=8),
        "eax": RegisterAliasDef(name="eax", parent="rax", size=4, offset=0),
        "ax": RegisterAliasDef(name="ax", parent="rax", size=2, offset=0),
        "al": RegisterAliasDef(name="al", parent="rax", size=1, offset=0),
        "ah": RegisterAliasDef(name="ah", parent="rax", size=1, offset=1),
        "rbx": RegisterDef(name="rbx", size=8),
        "ebx": RegisterAliasDef(name="ebx", parent="rbx", size=4, offset=0),
        "bx": RegisterAliasDef(name="bx", parent="rbx", size=2, offset=0),
        "bl": RegisterAliasDef(name="bl", parent="rbx", size=1, offset=0),
        "bh": RegisterAliasDef(name="bh", parent="rbx", size=1, offset=1),
        "rcx": RegisterDef(name="rcx", size=8),
        "ecx": RegisterAliasDef(name="ecx", parent="rcx", size=4, offset=0),
        "cx": RegisterAliasDef(name="cx", parent="rcx", size=2, offset=0),
        "cl": RegisterAliasDef(name="cl", parent="rcx", size=1, offset=0),
        "ch": RegisterAliasDef(name="ch", parent="rcx", size=1, offset=1),
        "rdx": RegisterDef(name="rdx", size=8),
        "edx": RegisterAliasDef(name="edx", parent="rdx", size=4, offset=0),
        "dx": RegisterAliasDef(name="dx", parent="rdx", size=2, offset=0),
        "dl": RegisterAliasDef(name="dl", parent="rdx", size=1, offset=0),
        "dh": RegisterAliasDef(name="dh", parent="rdx", size=1, offset=1),
        "r8": RegisterDef(name="r8", size=8),
        "r8d": RegisterAliasDef(name="r8d", parent="r8", size=4, offset=0),
        "r8w": RegisterAliasDef(name="r8w", parent="r8", size=2, offset=0),
        "r8b": RegisterAliasDef(name="r8b", parent="r8", size=1, offset=0),
        "r9": RegisterDef(name="r9", size=8),
        "r9d": RegisterAliasDef(name="r9d", parent="r9", size=4, offset=0),
        "r9w": RegisterAliasDef(name="r9w", parent="r9", size=2, offset=0),
        "r9b": RegisterAliasDef(name="r9b", parent="r9", size=1, offset=0),
        "r10": RegisterDef(name="r10", size=8),
        "r10d": RegisterAliasDef(name="r10d", parent="r10", size=4, offset=0),
        "r10w": RegisterAliasDef(name="r10w", parent="r10", size=2, offset=0),
        "r10b": RegisterAliasDef(name="r10b", parent="r10", size=1, offset=0),
        "r11": RegisterDef(name="r11", size=8),
        "r11d": RegisterAliasDef(name="r11d", parent="r11", size=4, offset=0),
        "r11w": RegisterAliasDef(name="r11w", parent="r11", size=2, offset=0),
        "r11b": RegisterAliasDef(name="r11b", parent="r11", size=1, offset=0),
        "r12": RegisterDef(name="r12", size=8),
        "r12d": RegisterAliasDef(name="r12d", parent="r12", size=4, offset=0),
        "r12w": RegisterAliasDef(name="r12w", parent="r12", size=2, offset=0),
        "r12b": RegisterAliasDef(name="r12b", parent="r12", size=1, offset=0),
        "r13": RegisterDef(name="r13", size=8),
        "r13d": RegisterAliasDef(name="r13d", parent="r13", size=4, offset=0),
        "r13w": RegisterAliasDef(name="r13w", parent="r13", size=2, offset=0),
        "r13b": RegisterAliasDef(name="r13b", parent="r13", size=1, offset=0),
        "r14": RegisterDef(name="r14", size=8),
        "r14d": RegisterAliasDef(name="r14d", parent="r14", size=4, offset=0),
        "r14w": RegisterAliasDef(name="r14w", parent="r14", size=2, offset=0),
        "r14b": RegisterAliasDef(name="r14b", parent="r14", size=1, offset=0),
        "r15": RegisterDef(name="r15", size=8),
        "r15d": RegisterAliasDef(name="r15d", parent="r15", size=4, offset=0),
        "r15w": RegisterAliasDef(name="r15w", parent="r15", size=2, offset=0),
        "r15b": RegisterAliasDef(name="r15b", parent="r15", size=1, offset=0),
        "rdi": RegisterDef(name="rdi", size=8),
        "edi": RegisterAliasDef(name="edi", parent="rdi", size=4, offset=0),
        "di": RegisterAliasDef(name="di", parent="rdi", size=2, offset=0),
        "dil": RegisterAliasDef(name="dil", parent="rdi", size=1, offset=0),
        "rsi": RegisterDef(name="rsi", size=8),
        "esi": RegisterAliasDef(name="esi", parent="rsi", size=4, offset=0),
        "si": RegisterAliasDef(name="si", parent="rsi", size=2, offset=0),
        "sil": RegisterAliasDef(name="sil", parent="rsi", size=1, offset=0),
        "rsp": RegisterDef(name="rsp", size=8),
        "esp": RegisterAliasDef(name="esp", parent="rsp", size=4, offset=0),
        "sp": RegisterAliasDef(name="sp", parent="rsp", size=2, offset=0),
        "spl": RegisterAliasDef(name="spl", parent="rsp", size=1, offset=0),
        "rbp": RegisterDef(name="rbp", size=8),
        "ebp": RegisterAliasDef(name="ebp", parent="rbp", size=4, offset=0),
        "bp": RegisterAliasDef(name="bp", parent="rbp", size=2, offset=0),
        "bpl": RegisterAliasDef(name="bpl", parent="rbp", size=1, offset=0),
        # *** Instruction Pointer ***
        "rip": RegisterDef(name="rip", size=8),
        "eip": RegisterAliasDef(name="eip", parent="rip", size=4, offset=0),
        "ip": RegisterAliasDef(name="ip", parent="rip", size=2, offset=0),
        "pc": RegisterAliasDef(name="pc", parent="rip", size=8, offset=0),
        # *** Flags register ***
        "rflags": RegisterDef(name="rflags", size=8),
        "eflags": RegisterAliasDef(name="eflags", parent="rflags", size=4, offset=0),
        "flags": RegisterAliasDef(name="flags", parent="rflags", size=2, offset=0),
        # *** Segment Registers ***
        # NOTE: These are actually 16 bits
        # However, their representation in different emulators gets weird.
        "cs": RegisterDef(name="cs", size=8),
        "ds": RegisterDef(name="ds", size=8),
        "es": RegisterDef(name="es", size=8),
        "fs": RegisterDef(name="fs", size=8),
        "gs": RegisterDef(name="gs", size=8),
        "ss": RegisterDef(name="ss", size=8),
        # *** Control Registers ***
        "cr0": RegisterDef(name="cr0", size=8),
        "cr1": RegisterDef(name="cr1", size=8),
        "cr2": RegisterDef(name="cr2", size=8),
        "cr3": RegisterDef(name="cr3", size=8),
        "cr4": RegisterDef(name="cr4", size=8),
        "cr8": RegisterDef(name="cr8", size=8),
        # *** Debug Registers ***
        "dr0": RegisterDef(name="dr0", size=8),
        "dr1": RegisterDef(name="dr1", size=8),
        "dr2": RegisterDef(name="dr2", size=8),
        "dr3": RegisterDef(name="dr3", size=8),
        "dr6": RegisterDef(name="dr6", size=8),
        "dr7": RegisterDef(name="dr7", size=8),
        "dr8": RegisterDef(name="dr8", size=8),
        "dr9": RegisterDef(name="dr9", size=8),
        "dr10": RegisterDef(name="dr10", size=8),
        "dr11": RegisterDef(name="dr11", size=8),
        "dr12": RegisterDef(name="dr12", size=8),
        "dr13": RegisterDef(name="dr13", size=8),
        "dr14": RegisterDef(name="dr14", size=8),
        "dr15": RegisterDef(name="dr15", size=8),
        # *** Descriptor Table Registers ***
        "gdtr": RegisterDef(name="gdtr", size=10),
        "idtr": RegisterDef(name="idtr", size=10),
        "ldtr": RegisterDef(name="ldtr", size=10),
        # *** Task Register ***
        "tr": RegisterDef(name="tr", size=2),
        # *** x87 registers ***
        # NOTE: These represent the physical registers.
        # x87 references registers as a stack,
        # which isn't possible to represent statically.
        "fpr0": RegisterDef(name="fpr0", size=10),
        "fpr1": RegisterDef(name="fpr1", size=10),
        "fpr2": RegisterDef(name="fpr2", size=10),
        "fpr3": RegisterDef(name="fpr3", size=10),
        "fpr4": RegisterDef(name="fpr4", size=10),
        "fpr5": RegisterDef(name="fpr5", size=10),
        "fpr6": RegisterDef(name="fpr6", size=10),
        "fpr7": RegisterDef(name="fpr7", size=10),
        # x87 Control Register
        "fctrl": RegisterDef(name="fctrl", size=2),
        # x87 Status Register
        "fstat": RegisterDef(name="fstat", size=2),
        # x87 Tag Register
        "ftag": RegisterDef(name="ftag", size=2),
        # x87 Last Instruction Register
        "fip": RegisterDef(name="fip", size=8),
        # x87 Last Operand Pointer
        "fdp": RegisterDef(name="fdp", size=8),
        # x87 Last Opcode
        "fop": RegisterDef(name="fop", size=2),
        # NOTE: Docs disagree on the format of fip and fdp.
        # One source describes them as 48-bit offset-plus-segment,
        # the other describes them as 64-bit.
        # There may also be separate segment registers.
        # If you care about the x87 debug info, please feel free to update.
        # *** MMX Registers ***
        # NOTE: The MMX registers are aliases for the low 8 bytes of the x87 registers.
        # The two subsystems cannot be used simultaneously.
        "mm0": RegisterAliasDef(name="mm0", parent="fpr0", size=8, offset=0),
        "mm1": RegisterAliasDef(name="mm1", parent="fpr1", size=8, offset=0),
        "mm2": RegisterAliasDef(name="mm2", parent="fpr2", size=8, offset=0),
        "mm3": RegisterAliasDef(name="mm3", parent="fpr3", size=8, offset=0),
        "mm4": RegisterAliasDef(name="mm4", parent="fpr4", size=8, offset=0),
        "mm5": RegisterAliasDef(name="mm5", parent="fpr5", size=8, offset=0),
        "mm6": RegisterAliasDef(name="mm6", parent="fpr6", size=8, offset=0),
        "mm7": RegisterAliasDef(name="mm7", parent="fpr7", size=8, offset=0),
    }


class AMD64(AMD64BasePlatformDef):
    architecture = Architecture.X86_64

    registers = AMD64BasePlatformDef.registers | {
        # *** SSE/AVX/AVX2 registers ***
        "ymm0": RegisterDef(name="ymm0", size=32),
        "xmm0": RegisterAliasDef(name="xmm0", parent="ymm0", size=16, offset=0),
        "ymm1": RegisterDef(name="ymm1", size=32),
        "xmm1": RegisterAliasDef(name="xmm1", parent="ymm1", size=16, offset=0),
        "ymm2": RegisterDef(name="ymm2", size=32),
        "xmm2": RegisterAliasDef(name="xmm2", parent="ymm2", size=16, offset=0),
        "ymm3": RegisterDef(name="ymm3", size=32),
        "xmm3": RegisterAliasDef(name="xmm3", parent="ymm3", size=16, offset=0),
        "ymm4": RegisterDef(name="ymm4", size=32),
        "xmm4": RegisterAliasDef(name="xmm4", parent="ymm4", size=16, offset=0),
        "ymm5": RegisterDef(name="ymm5", size=32),
        "xmm5": RegisterAliasDef(name="xmm5", parent="ymm5", size=16, offset=0),
        "ymm6": RegisterDef(name="ymm6", size=32),
        "xmm6": RegisterAliasDef(name="xmm6", parent="ymm6", size=16, offset=0),
        "ymm7": RegisterDef(name="ymm7", size=32),
        "xmm7": RegisterAliasDef(name="xmm7", parent="ymm7", size=16, offset=0),
        "ymm8": RegisterDef(name="ymm8", size=32),
        "xmm8": RegisterAliasDef(name="xmm8", parent="ymm8", size=16, offset=0),
        "ymm9": RegisterDef(name="ymm9", size=32),
        "xmm9": RegisterAliasDef(name="xmm9", parent="ymm9", size=16, offset=0),
        "ymm10": RegisterDef(name="ymm10", size=32),
        "xmm10": RegisterAliasDef(name="xmm10", parent="ymm10", size=16, offset=0),
        "ymm11": RegisterDef(name="ymm11", size=32),
        "xmm11": RegisterAliasDef(name="xmm11", parent="ymm11", size=16, offset=0),
        "ymm12": RegisterDef(name="ymm12", size=32),
        "xmm12": RegisterAliasDef(name="xmm12", parent="ymm12", size=16, offset=0),
        "ymm13": RegisterDef(name="ymm13", size=32),
        "xmm13": RegisterAliasDef(name="xmm13", parent="ymm13", size=16, offset=0),
        "ymm14": RegisterDef(name="ymm14", size=32),
        "xmm14": RegisterAliasDef(name="xmm14", parent="ymm14", size=16, offset=0),
        "ymm15": RegisterDef(name="ymm15", size=32),
        "xmm15": RegisterAliasDef(name="xmm15", parent="ymm15", size=16, offset=0),
    }


class AMD64AVX512(AMD64BasePlatformDef):
    architecture = Architecture.X86_64_AVX512

    registers = AMD64BasePlatformDef.registers | {
        # *** SSE/AVX/AVX2/AVX512 registers ***
        "zmm0": RegisterDef(name="zmm0", size=64),
        "ymm0": RegisterAliasDef(name="ymm0", parent="zmm0", size=32, offset=0),
        "xmm0": RegisterAliasDef(name="xmm0", parent="zmm0", size=16, offset=0),
        "zmm1": RegisterDef(name="zmm1", size=64),
        "ymm1": RegisterAliasDef(name="ymm1", parent="zmm1", size=32, offset=0),
        "xmm1": RegisterAliasDef(name="xmm1", parent="zmm1", size=16, offset=0),
        "zmm2": RegisterDef(name="zmm2", size=64),
        "ymm2": RegisterAliasDef(name="ymm2", parent="zmm2", size=32, offset=0),
        "xmm2": RegisterAliasDef(name="xmm2", parent="zmm2", size=16, offset=0),
        "zmm3": RegisterDef(name="zmm3", size=64),
        "ymm3": RegisterAliasDef(name="ymm3", parent="zmm3", size=32, offset=0),
        "xmm3": RegisterAliasDef(name="xmm3", parent="zmm3", size=16, offset=0),
        "zmm4": RegisterDef(name="zmm4", size=64),
        "ymm4": RegisterAliasDef(name="ymm4", parent="zmm4", size=32, offset=0),
        "xmm4": RegisterAliasDef(name="xmm4", parent="zmm4", size=16, offset=0),
        "zmm5": RegisterDef(name="zmm5", size=64),
        "ymm5": RegisterAliasDef(name="ymm5", parent="zmm5", size=32, offset=0),
        "xmm5": RegisterAliasDef(name="xmm5", parent="zmm5", size=16, offset=0),
        "zmm6": RegisterDef(name="zmm6", size=64),
        "ymm6": RegisterAliasDef(name="ymm6", parent="zmm6", size=32, offset=0),
        "xmm6": RegisterAliasDef(name="xmm6", parent="zmm6", size=16, offset=0),
        "zmm7": RegisterDef(name="zmm7", size=64),
        "ymm7": RegisterAliasDef(name="ymm7", parent="zmm7", size=32, offset=0),
        "xmm7": RegisterAliasDef(name="xmm7", parent="zmm7", size=16, offset=0),
        "zmm8": RegisterDef(name="zmm8", size=64),
        "ymm8": RegisterAliasDef(name="ymm8", parent="zmm8", size=32, offset=0),
        "xmm8": RegisterAliasDef(name="xmm8", parent="zmm8", size=16, offset=0),
        "zmm9": RegisterDef(name="zmm9", size=64),
        "ymm9": RegisterAliasDef(name="ymm9", parent="zmm9", size=32, offset=0),
        "xmm9": RegisterAliasDef(name="xmm9", parent="zmm9", size=16, offset=0),
        "zmm10": RegisterDef(name="zmm10", size=64),
        "ymm10": RegisterAliasDef(name="ymm10", parent="zmm10", size=32, offset=0),
        "xmm10": RegisterAliasDef(name="xmm10", parent="zmm10", size=16, offset=0),
        "zmm11": RegisterDef(name="zmm11", size=64),
        "ymm11": RegisterAliasDef(name="ymm11", parent="zmm11", size=32, offset=0),
        "xmm11": RegisterAliasDef(name="xmm11", parent="zmm11", size=16, offset=0),
        "zmm12": RegisterDef(name="zmm12", size=64),
        "ymm12": RegisterAliasDef(name="ymm12", parent="zmm12", size=32, offset=0),
        "xmm12": RegisterAliasDef(name="xmm12", parent="zmm12", size=16, offset=0),
        "zmm13": RegisterDef(name="zmm13", size=64),
        "ymm13": RegisterAliasDef(name="ymm13", parent="zmm13", size=32, offset=0),
        "xmm13": RegisterAliasDef(name="xmm13", parent="zmm13", size=16, offset=0),
        "zmm14": RegisterDef(name="zmm14", size=64),
        "ymm14": RegisterAliasDef(name="ymm14", parent="zmm14", size=32, offset=0),
        "xmm14": RegisterAliasDef(name="xmm14", parent="zmm14", size=16, offset=0),
        "zmm15": RegisterDef(name="zmm15", size=64),
        "ymm15": RegisterAliasDef(name="ymm15", parent="zmm15", size=32, offset=0),
        "xmm15": RegisterAliasDef(name="xmm15", parent="zmm15", size=16, offset=0),
        "zmm16": RegisterDef(name="zmm16", size=64),
        "ymm16": RegisterAliasDef(name="ymm16", parent="zmm16", size=32, offset=0),
        "xmm16": RegisterAliasDef(name="xmm16", parent="zmm16", size=16, offset=0),
        "zmm17": RegisterDef(name="zmm17", size=64),
        "ymm17": RegisterAliasDef(name="ymm17", parent="zmm17", size=32, offset=0),
        "xmm17": RegisterAliasDef(name="xmm17", parent="zmm17", size=16, offset=0),
        "zmm18": RegisterDef(name="zmm18", size=64),
        "ymm18": RegisterAliasDef(name="ymm18", parent="zmm18", size=32, offset=0),
        "xmm18": RegisterAliasDef(name="xmm18", parent="zmm18", size=16, offset=0),
        "zmm19": RegisterDef(name="zmm19", size=64),
        "ymm19": RegisterAliasDef(name="ymm19", parent="zmm19", size=32, offset=0),
        "xmm19": RegisterAliasDef(name="xmm19", parent="zmm19", size=16, offset=0),
        "zmm20": RegisterDef(name="zmm20", size=64),
        "ymm20": RegisterAliasDef(name="ymm20", parent="zmm20", size=32, offset=0),
        "xmm20": RegisterAliasDef(name="xmm20", parent="zmm20", size=16, offset=0),
        "zmm21": RegisterDef(name="zmm21", size=64),
        "ymm21": RegisterAliasDef(name="ymm21", parent="zmm21", size=32, offset=0),
        "xmm21": RegisterAliasDef(name="xmm21", parent="zmm21", size=16, offset=0),
        "zmm22": RegisterDef(name="zmm22", size=64),
        "ymm22": RegisterAliasDef(name="ymm22", parent="zmm22", size=32, offset=0),
        "xmm22": RegisterAliasDef(name="xmm22", parent="zmm22", size=16, offset=0),
        "zmm23": RegisterDef(name="zmm23", size=64),
        "ymm23": RegisterAliasDef(name="ymm23", parent="zmm23", size=32, offset=0),
        "xmm23": RegisterAliasDef(name="xmm23", parent="zmm23", size=16, offset=0),
        "zmm24": RegisterDef(name="zmm24", size=64),
        "ymm24": RegisterAliasDef(name="ymm24", parent="zmm24", size=32, offset=0),
        "xmm24": RegisterAliasDef(name="xmm24", parent="zmm24", size=16, offset=0),
        "zmm25": RegisterDef(name="zmm25", size=64),
        "ymm25": RegisterAliasDef(name="ymm25", parent="zmm25", size=32, offset=0),
        "xmm25": RegisterAliasDef(name="xmm25", parent="zmm25", size=16, offset=0),
        "zmm26": RegisterDef(name="zmm26", size=64),
        "ymm26": RegisterAliasDef(name="ymm26", parent="zmm26", size=32, offset=0),
        "xmm26": RegisterAliasDef(name="xmm26", parent="zmm26", size=16, offset=0),
        "zmm27": RegisterDef(name="zmm27", size=64),
        "ymm27": RegisterAliasDef(name="ymm27", parent="zmm27", size=32, offset=0),
        "xmm27": RegisterAliasDef(name="xmm27", parent="zmm27", size=16, offset=0),
        "zmm28": RegisterDef(name="zmm28", size=64),
        "ymm28": RegisterAliasDef(name="ymm28", parent="zmm28", size=32, offset=0),
        "xmm28": RegisterAliasDef(name="xmm28", parent="zmm28", size=16, offset=0),
        "zmm29": RegisterDef(name="zmm29", size=64),
        "ymm29": RegisterAliasDef(name="ymm29", parent="zmm29", size=32, offset=0),
        "xmm29": RegisterAliasDef(name="xmm29", parent="zmm29", size=16, offset=0),
        "zmm30": RegisterDef(name="zmm30", size=64),
        "ymm30": RegisterAliasDef(name="ymm30", parent="zmm30", size=32, offset=0),
        "xmm30": RegisterAliasDef(name="xmm30", parent="zmm30", size=16, offset=0),
        "zmm31": RegisterDef(name="zmm31", size=64),
        "ymm31": RegisterAliasDef(name="ymm31", parent="zmm31", size=32, offset=0),
        "xmm31": RegisterAliasDef(name="xmm31", parent="zmm31", size=16, offset=0),
    }



from ....c99 import (
    Asctime,
    Clock,
    Ctime,
    Difftime,
    Gmtime,
    Localtime,
    Mktime,
    Strftime,
    Time,
)
from ..systemv import ArmHFSysVModel


class ArmHFSysVTime(Time, ArmHFSysVModel):
    pass


class ArmHFSysVLocaltime(Localtime, ArmHFSysVModel):
    pass


class ArmHFSysVGmtime(Gmtime, ArmHFSysVModel):
    pass


class ArmHFSysVCtime(Ctime, ArmHFSysVModel):
    pass


class ArmHFSysVAsctime(Asctime, ArmHFSysVModel):
    pass


class ArmHFSysVStrftime(Strftime, ArmHFSysVModel):
    pass


class ArmHFSysVDifftime(Difftime, ArmHFSysVModel):
    pass


class ArmHFSysVMktime(Mktime, ArmHFSysVModel):
    pass


class ArmHFSysVClock(Clock, ArmHFSysVModel):
    pass


__all__ = [
    "ArmHFSysVTime",
    "ArmHFSysVLocaltime",
    "ArmHFSysVGmtime",
    "ArmHFSysVCtime",
    "ArmHFSysVAsctime",
    "ArmHFSysVStrftime",
    "ArmHFSysVDifftime",
    "ArmHFSysVMktime",
    "ArmHFSysVClock",
]



from ....c99 import (
    Abort,
    Abs,
    Atexit,
    Atof,
    Atoi,
    Atol,
    Atoll,
    Calloc,
    Div,
    Exit,
    Free,
    Getenv,
    LAbs,
    LDiv,
    LLAbs,
    LLDiv,
    Malloc,
    Mblen,
    Mbstowcs,
    Mbtowc,
    QSort,
    Rand,
    Realloc,
    Srand,
    System,
    Wcstombs,
    Wctomb,
)
from ..systemv import ArmHFSysVModel


class ArmHFSysVAbs(Abs, ArmHFSysVModel):
    pass


class ArmHFSysVLAbs(LAbs, ArmHFSysVModel):
    pass


class ArmHFSysVLLAbs(LLAbs, ArmHFSysVModel):
    pass


class ArmHFSysVAtof(Atof, ArmHFSysVModel):
    pass


class ArmHFSysVAtoi(Atoi, ArmHFSysVModel):
    pass


class ArmHFSysVAtol(Atol, ArmHFSysVModel):
    pass


class ArmHFSysVAtoll(Atoll, ArmHFSysVModel):
    pass


class ArmHFSysVCalloc(Calloc, ArmHFSysVModel):
    pass


class ArmHFSysVDiv(Div, ArmHFSysVModel):
    pass


class ArmHFSysVLDiv(LDiv, ArmHFSysVModel):
    pass


class ArmHFSysVLLDiv(LLDiv, ArmHFSysVModel):
    pass


class ArmHFSysVExit(Exit, ArmHFSysVModel):
    pass


class ArmHFSysVFree(Free, ArmHFSysVModel):
    pass


class ArmHFSysVMalloc(Malloc, ArmHFSysVModel):
    pass


class ArmHFSysVQSort(QSort, ArmHFSysVModel):
    pass


class ArmHFSysVRand(Rand, ArmHFSysVModel):
    pass


class ArmHFSysVRealloc(Realloc, ArmHFSysVModel):
    pass


class ArmHFSysVSrand(Srand, ArmHFSysVModel):
    pass


class ArmHFSysVAbort(Abort, ArmHFSysVModel):
    pass


class ArmHFSysVAtexit(Atexit, ArmHFSysVModel):
    pass


class ArmHFSysVGetenv(Getenv, ArmHFSysVModel):
    pass


class ArmHFSysVMblen(Mblen, ArmHFSysVModel):
    pass


class ArmHFSysVMbstowcs(Mbstowcs, ArmHFSysVModel):
    pass


class ArmHFSysVMbtowc(Mbtowc, ArmHFSysVModel):
    pass


class ArmHFSysVSystem(System, ArmHFSysVModel):
    pass


class ArmHFSysVWcstombs(Wcstombs, ArmHFSysVModel):
    pass


class ArmHFSysVWctomb(Wctomb, ArmHFSysVModel):
    pass


__all__ = [
    "ArmHFSysVAbs",
    "ArmHFSysVLAbs",
    "ArmHFSysVLLAbs",
    "ArmHFSysVAtof",
    "ArmHFSysVAtoi",
    "ArmHFSysVAtol",
    "ArmHFSysVAtoll",
    "ArmHFSysVCalloc",
    "ArmHFSysVDiv",
    "ArmHFSysVLDiv",
    "ArmHFSysVLLDiv",
    "ArmHFSysVExit",
    "ArmHFSysVFree",
    "ArmHFSysVMalloc",
    "ArmHFSysVQSort",
    "ArmHFSysVRand",
    "ArmHFSysVRealloc",
    "ArmHFSysVSrand",
    "ArmHFSysVAbort",
    "ArmHFSysVAtexit",
    "ArmHFSysVGetenv",
    "ArmHFSysVMblen",
    "ArmHFSysVMbstowcs",
    "ArmHFSysVMbtowc",
    "ArmHFSysVSystem",
    "ArmHFSysVWcstombs",
    "ArmHFSysVWctomb",
]



import typing

import claripy

from ... import emulators
from ...platforms import Platform
from ..state import Value
from . import memory


class Executable(memory.RawMemory):
    """An execuable piece of code."""

    entrypoint: typing.Optional[int] = None
    """Entry point address.

    Note:
        This is not used by Emulators - but is available for reference from
        file parsing, if supported.
    """

    def _write_content(self, emulator: emulators.Emulator, address: int, value: Value):
        if isinstance(value.get_content(), claripy.ast.bv.BV):
            raise NotImplementedError(
                "Absolutely not.  Get your symbolic code out of here."
            )
        emulator.write_code(address, value.to_bytes(emulator.platform.byteorder))

    @classmethod
    def from_elf(
        cls,
        file: typing.BinaryIO,
        address: typing.Optional[int] = None,
        platform: typing.Optional[Platform] = None,
        ignore_platform: bool = False,
        page_size: int = 0x1000,
    ):
        """Load an ELF executable from an open file-like object.

        Arguments:
            file: The open file-like object from which to read.
            address: The address where this executable should be loaded.
            platform: Optional platform for header verification
            ignore_platform: Skip platform ID and verification
            page_size: Page size in bytes

        Returns:
            An Executable parsed from the given ELF file-like object.
        """
        # NOTE: there's a circular dependency between elf.py and code.py
        # This is the accepted fix.
        from .elf import ElfExecutable

        return ElfExecutable(
            file,
            user_base=address,
            platform=platform,
            ignore_platform=ignore_platform,
            page_size=page_size,
        )

    @classmethod
    def from_elf_core(
        cls,
        file: typing.BinaryIO,
        address: typing.Optional[int] = None,
        platform: typing.Optional[Platform] = None,
        ignore_platform: bool = False,
    ):
        """
        Load an ELF core dump (ET_CORE) from an open file-like object.

        Arguments:
            file: The open file-like object from which to read.
            address: The address where this core dump should be loaded.
            platform: Optional platform for header verification
            ignore_platform: Skip platform ID and verification

        Returns:
            An Executable (specifically ElfCoreFile) parsed from the given core dump.
        """
        from .elf import ElfCoreFile

        return ElfCoreFile(
            file, user_base=address, platform=platform, ignore_platform=ignore_platform
        )

    @classmethod
    def from_pe(
        cls,
        file: typing.BinaryIO,
        address: typing.Optional[int] = None,
        platform: typing.Optional[Platform] = None,
        ignore_platform: bool = False,
    ):
        """Load an PE executable from an open file-like object.

        Arguments:
            file: The open file-like object from which to read.
            address: The address where this executable should be loaded.

        Returns:
            An Executable parsed from the given PE file-like object.
        """
        from .pe import PEExecutable

        return PEExecutable(
            file, user_base=address, platform=platform, ignore_platform=ignore_platform
        )


__all__ = ["Executable"]



from ....c99 import (
    Asctime,
    Clock,
    Ctime,
    Difftime,
    Gmtime,
    Localtime,
    Mktime,
    Strftime,
    Time,
)
from ..systemv import AMD64SysVModel


class AMD64SysVTime(Time, AMD64SysVModel):
    pass


class AMD64SysVLocaltime(Localtime, AMD64SysVModel):
    pass


class AMD64SysVGmtime(Gmtime, AMD64SysVModel):
    pass


class AMD64SysVCtime(Ctime, AMD64SysVModel):
    pass


class AMD64SysVAsctime(Asctime, AMD64SysVModel):
    pass


class AMD64SysVStrftime(Strftime, AMD64SysVModel):
    pass


class AMD64SysVDifftime(Difftime, AMD64SysVModel):
    pass


class AMD64SysVMktime(Mktime, AMD64SysVModel):
    pass


class AMD64SysVClock(Clock, AMD64SysVModel):
    pass


__all__ = [
    "AMD64SysVTime",
    "AMD64SysVLocaltime",
    "AMD64SysVGmtime",
    "AMD64SysVCtime",
    "AMD64SysVAsctime",
    "AMD64SysVStrftime",
    "AMD64SysVDifftime",
    "AMD64SysVMktime",
    "AMD64SysVClock",
]



from ...... import platforms
from .prstatus import PrStatus


class I386(PrStatus):
    architecture = platforms.Architecture.X86_32
    byteorder = platforms.Byteorder.LITTLE

    pr_regs_off = 72
    pr_regs_size = 68

    register_coords = [
        ("ebx", 0x0, 4),
        ("ecx", 0x4, 4),
        ("edx", 0x8, 4),
        ("esi", 0xC, 4),
        ("edi", 0x10, 4),
        ("ebp", 0x14, 4),
        ("eax", 0x18, 4),
        ("ds", 0x1C, 4),
        ("es", 0x20, 4),
        ("fs", 0x24, 4),
        ("gs", 0x28, 4),
        (None, 0x2C, 4),  # orig_eax
        ("eip", 0x30, 4),
        ("cs", 0x34, 4),
        ("eflags", 0x38, 4),
        ("esp", 0x3C, 4),
        ("ss", 0x40, 4),
    ]



from ....c99 import (
    Clearerr,
    Fclose,
    Feof,
    Ferror,
    Fflush,
    Fgetc,
    Fgetpos,
    Fgets,
    Fopen,
    Fprintf,
    Fputc,
    Fputs,
    Fread,
    Freopen,
    Fscanf,
    Fseek,
    Fsetpos,
    Ftell,
    Fwrite,
    Getc,
    Getchar,
    Gets,
    Printf,
    Putc,
    Putchar,
    Puts,
    Remove,
    Rename,
    Rewind,
    Scanf,
    Snprintf,
    Sprintf,
    Sscanf,
    Tmpfile,
    Tmpnam,
    Ungetc,
    Vfprintf,
    Vfscanf,
    Vprintf,
    Vscanf,
    Vsnprintf,
    Vsprintf,
    Vsscanf,
)
from ..systemv import MIPS64SysVModel


class MIPS64SysVFclose(Fclose, MIPS64SysVModel):
    pass


class MIPS64SysVFeof(Feof, MIPS64SysVModel):
    pass


class MIPS64SysVFerror(Ferror, MIPS64SysVModel):
    pass


class MIPS64SysVFgetc(Fgetc, MIPS64SysVModel):
    pass


class MIPS64SysVFgets(Fgets, MIPS64SysVModel):
    pass


class MIPS64SysVFopen(Fopen, MIPS64SysVModel):
    pass


class MIPS64SysVFprintf(Fprintf, MIPS64SysVModel):
    pass


class MIPS64SysVFputc(Fputc, MIPS64SysVModel):
    pass


class MIPS64SysVFputs(Fputs, MIPS64SysVModel):
    pass


class MIPS64SysVFread(Fread, MIPS64SysVModel):
    pass


class MIPS64SysVFscanf(Fscanf, MIPS64SysVModel):
    pass


class MIPS64SysVFseek(Fseek, MIPS64SysVModel):
    pass


class MIPS64SysVFtell(Ftell, MIPS64SysVModel):
    pass


class MIPS64SysVFwrite(Fwrite, MIPS64SysVModel):
    pass


class MIPS64SysVGetc(Getc, MIPS64SysVModel):
    pass


class MIPS64SysVGetchar(Getchar, MIPS64SysVModel):
    pass


class MIPS64SysVGets(Gets, MIPS64SysVModel):
    pass


class MIPS64SysVPrintf(Printf, MIPS64SysVModel):
    pass


class MIPS64SysVPutc(Putc, MIPS64SysVModel):
    pass


class MIPS64SysVPutchar(Putchar, MIPS64SysVModel):
    pass


class MIPS64SysVPuts(Puts, MIPS64SysVModel):
    pass


class MIPS64SysVRemove(Remove, MIPS64SysVModel):
    pass


class MIPS64SysVRename(Rename, MIPS64SysVModel):
    pass


class MIPS64SysVRewind(Rewind, MIPS64SysVModel):
    pass


class MIPS64SysVScanf(Scanf, MIPS64SysVModel):
    pass


class MIPS64SysVSnprintf(Snprintf, MIPS64SysVModel):
    pass


class MIPS64SysVSprintf(Sprintf, MIPS64SysVModel):
    pass


class MIPS64SysVSscanf(Sscanf, MIPS64SysVModel):
    pass


class MIPS64SysVClearerr(Clearerr, MIPS64SysVModel):
    pass


class MIPS64SysVFflush(Fflush, MIPS64SysVModel):
    pass


class MIPS64SysVFreopen(Freopen, MIPS64SysVModel):
    pass


class MIPS64SysVFgetpos(Fgetpos, MIPS64SysVModel):
    pass


class MIPS64SysVFsetpos(Fsetpos, MIPS64SysVModel):
    pass


class MIPS64SysVTmpfile(Tmpfile, MIPS64SysVModel):
    pass


class MIPS64SysVTmpnam(Tmpnam, MIPS64SysVModel):
    pass


class MIPS64SysVUngetc(Ungetc, MIPS64SysVModel):
    pass


class MIPS64SysVVfprintf(Vfprintf, MIPS64SysVModel):
    pass


class MIPS64SysVVfscanf(Vfscanf, MIPS64SysVModel):
    pass


class MIPS64SysVVprintf(Vprintf, MIPS64SysVModel):
    pass


class MIPS64SysVVscanf(Vscanf, MIPS64SysVModel):
    pass


class MIPS64SysVVsnprintf(Vsnprintf, MIPS64SysVModel):
    pass


class MIPS64SysVVsprintf(Vsprintf, MIPS64SysVModel):
    pass


class MIPS64SysVVsscanf(Vsscanf, MIPS64SysVModel):
    pass


__all__ = [
    "MIPS64SysVFclose",
    "MIPS64SysVFeof",
    "MIPS64SysVFerror",
    "MIPS64SysVFgetc",
    "MIPS64SysVFgets",
    "MIPS64SysVFopen",
    "MIPS64SysVFprintf",
    "MIPS64SysVFputc",
    "MIPS64SysVFputs",
    "MIPS64SysVFread",
    "MIPS64SysVFscanf",
    "MIPS64SysVFseek",
    "MIPS64SysVFtell",
    "MIPS64SysVFwrite",
    "MIPS64SysVGetc",
    "MIPS64SysVGetchar",
    "MIPS64SysVGets",
    "MIPS64SysVPrintf",
    "MIPS64SysVPutc",
    "MIPS64SysVPutchar",
    "MIPS64SysVPuts",
    "MIPS64SysVRemove",
    "MIPS64SysVRename",
    "MIPS64SysVRewind",
    "MIPS64SysVScanf",
    "MIPS64SysVSnprintf",
    "MIPS64SysVSprintf",
    "MIPS64SysVSscanf",
    "MIPS64SysVClearerr",
    "MIPS64SysVFflush",
    "MIPS64SysVFreopen",
    "MIPS64SysVFgetpos",
    "MIPS64SysVFsetpos",
    "MIPS64SysVTmpfile",
    "MIPS64SysVTmpnam",
    "MIPS64SysVUngetc",
    "MIPS64SysVVfprintf",
    "MIPS64SysVVfscanf",
    "MIPS64SysVVprintf",
    "MIPS64SysVVscanf",
    "MIPS64SysVVsnprintf",
    "MIPS64SysVVsprintf",
    "MIPS64SysVVsscanf",
]
